from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from prov.model import *

EX_NS = Namespace('ex', 'http://example.org/')
EX2_NS = Namespace('ex2', 'http://example2.org/')


class TestStatementsBase(object):
    """This is the base class for testing different PROV statements.
    It is not runnable and needs to be included in a subclass of
    RoundTripTestCase.
    """

    def new_document(self):
        return ProvDocument()

    def add_label(self, record):
        record.add_attributes(
            [('prov:label', Literal("hello"))]
        )

    def add_labels(self, record):
        record.add_attributes([
            ('prov:label', Literal("hello")),
            ('prov:label', Literal("bye", langtag="en")),
            ('prov:label', Literal("bonjour", langtag="fr"))
        ])

    def add_types(self, record):
        record.add_attributes([
            ('prov:type', 'a'),
            ('prov:type', 1),
            ('prov:type', 1.0),
            ('prov:type', True),
            ('prov:type', EX_NS['abc']),
            ('prov:type', datetime.datetime.now()),
            ('prov:type', Literal('http://boiled-egg.example.com',
                                  datatype=XSD_ANYURI)),
        ])

    def add_locations(self, record):
        record.add_attributes([
            ('prov:Location', "Southampton"),
            ('prov:Location', 1),
            ('prov:Location', 1.0),
            ('prov:Location', True),
            ('prov:Location', EX_NS['london']),
            ('prov:Location', datetime.datetime.now()),
            ('prov:Location', EX_NS.uri + "london"),
            ('prov:Location', Literal(2002, datatype=XSD['gYear'])),
        ])

    def add_value(self, record):
        record.add_attributes([
            ('prov:value', EX_NS['avalue'])
        ])

    def add_further_attributes(self, record):
        record.add_attributes([
            (EX_NS['tag1'], "hello"),
            (EX_NS['tag2'], "bye"),
            (EX2_NS['tag3'], "hi"),
            (EX_NS['tag1'], "hello\nover\nmore\nlines"),
        ])

    def add_further_attributes0(self, record):
        record.add_attributes([
            (EX_NS['tag1'], "hello"),
            (EX_NS['tag2'], "bye"),
            (EX_NS['tag2'], Literal("hola", langtag="es")),
            (EX2_NS['tag3'], "hi"),
            (EX_NS['tag'], 1),
            # Long on Python 2, just int on Python 3.
            (EX_NS['tag'], six.integer_types[-1](1)),
            (EX_NS['tag'], Literal(1, datatype=XSD_SHORT)),
            (EX_NS['tag'], Literal(1, datatype=XSD_DOUBLE)),
            (EX_NS['tag'], 1.0),
            (EX_NS['tag'], True),
            (EX_NS['tag'], EX_NS.uri + "southampton"),
        ])

        self.add_further_attributes_with_qnames(record)

    def add_further_attributes_with_qnames(self, record):
        record.add_attributes([
            (EX_NS['tag'], EX2_NS['newyork']),
            (EX_NS['tag'], EX_NS['london']),
        ])

    # TESTS

    # ENTITIES
    def test_entity_0(self):
        document = self.new_document()
        a = document.entity(EX_NS['e0'])
        a.add_attributes([
            (EX_NS['tag2'], Literal("guten tag", langtag="de")),
            ('prov:Location', "un llieu"),
            (PROV['Location'], 1),
            (PROV['Location'], 2.0),
            (PROV['Location'], EX_NS.uri + "london"),
        ])
        self.do_tests(document)

    def test_entity_1(self):
        document = self.new_document()
        document.entity(EX_NS['e1'])
        self.do_tests(document)

    def test_entity_2(self):
        document = self.new_document()
        a = document.entity(EX_NS['e2'])
        a.add_attributes([(PROV_LABEL, 'entity2')])
        self.do_tests(document)

    def test_entity_3(self):
        document = self.new_document()
        a = document.entity(EX_NS['e3'])
        a.add_attributes([(PROV_LABEL, 'entity3')])
        self.add_value(a)
        self.do_tests(document)

    def test_entity_4(self):
        document = self.new_document()
        a = document.entity(EX_NS['e4'])
        a.add_attributes([(PROV_LABEL, 'entity4')])
        self.add_labels(a)
        self.do_tests(document)

    def test_entity_5(self):
        document = self.new_document()
        a = document.entity(EX_NS['e5'])
        a.add_attributes([(PROV_LABEL, 'entity5')])
        self.add_types(a)
        self.do_tests(document)

    def test_entity_6(self):
        document = self.new_document()
        a = document.entity(EX_NS['e6'])
        a.add_attributes([(PROV_LABEL, 'entity6')])
        self.add_locations(a)
        self.do_tests(document)

    def test_entity_7(self):
        document = self.new_document()
        a = document.entity(EX_NS['e7'])
        a.add_attributes([(PROV_LABEL, 'entity7')])
        self.add_types(a)
        self.add_locations(a)
        self.add_labels(a)
        self.do_tests(document)

    def test_entity_8(self):
        document = self.new_document()
        a = document.entity(EX_NS['e8'])
        a.add_attributes([(PROV_LABEL, 'entity8')])
        self.add_types(a)
        self.add_types(a)
        self.add_locations(a)
        self.add_locations(a)
        self.add_labels(a)
        self.add_labels(a)
        self.do_tests(document)

    def test_entity_9(self):
        document = self.new_document()
        a = document.entity(EX_NS['e9'])
        a.add_attributes([(PROV_LABEL, 'entity9')])
        self.add_types(a)
        self.add_locations(a)
        self.add_labels(a)
        self.add_further_attributes(a)
        self.do_tests(document)

    def test_entity_10(self):
        document = self.new_document()
        a = document.entity(EX_NS['e10'])
        a.add_attributes([(PROV_LABEL, 'entity10')])
        self.add_types(a)
        self.add_locations(a)
        self.add_labels(a)
        self.add_further_attributes0(a)
        self.do_tests(document)

    # ACTIVITIES
    def test_activity_1(self):
        document = self.new_document()
        document.activity(EX_NS['a1'])
        self.do_tests(document)

    def test_activity_2(self):
        document = self.new_document()
        a = document.activity(EX_NS['a2'])
        a.add_attributes([(PROV_LABEL, 'activity2')])
        self.do_tests(document)

    def test_activity_3(self):
        document = self.new_document()
        document.activity(
            EX_NS['a3'],
            startTime=datetime.datetime.now(),
            endTime=datetime.datetime.now()
        )
        self.do_tests(document)

    def test_activity_4(self):
        document = self.new_document()
        a = document.activity(EX_NS['a4'])
        a.add_attributes([(PROV_LABEL, 'activity4')])
        self.add_labels(a)
        self.do_tests(document)

    def test_activity_5(self):
        document = self.new_document()
        a = document.activity(EX_NS['a5'])
        a.add_attributes([(PROV_LABEL, 'activity5')])
        self.add_types(a)
        self.do_tests(document)

    def test_activity_6(self):
        document = self.new_document()
        a = document.activity(EX_NS['a6'])
        a.add_attributes([(PROV_LABEL, 'activity6')])
        self.add_locations(a)
        self.do_tests(document)

    def test_activity_7(self):
        document = self.new_document()
        a = document.activity(EX_NS['a7'])
        a.add_attributes([(PROV_LABEL, 'activity7')])
        self.add_types(a)
        self.add_locations(a)
        self.add_labels(a)
        self.do_tests(document)

    def test_activity_8(self):
        document = self.new_document()
        a = document.activity(
            EX_NS['a8'],
            startTime=datetime.datetime.now(),
            endTime=datetime.datetime.now()
        )
        a.add_attributes([(PROV_LABEL, 'activity8')])
        self.add_types(a)
        self.add_types(a)
        self.add_locations(a)
        self.add_locations(a)
        self.add_labels(a)
        self.add_labels(a)
        self.do_tests(document)

    def test_activity_9(self):
        document = self.new_document()
        a = document.activity(EX_NS['a9'])
        a.add_attributes([(PROV_LABEL, 'activity9')])
        self.add_types(a)
        self.add_locations(a)
        self.add_labels(a)
        self.add_further_attributes(a)
        self.do_tests(document)

    # AGENTS
    def test_agent_1(self):
        document = self.new_document()
        document.agent(EX_NS['ag1'])
        self.do_tests(document)

    def test_agent_2(self):
        document = self.new_document()
        a = document.agent(EX_NS['ag2'])
        a.add_attributes([(PROV_LABEL, 'agent2')])
        self.do_tests(document)

    def test_agent_3(self):
        document = self.new_document()
        a = document.agent(EX_NS['ag3'])
        a.add_attributes([
            (PROV_LABEL, 'agent3'),
            (PROV_LABEL, Literal('hello')),
        ])
        self.do_tests(document)

    def test_agent_4(self):
        document = self.new_document()
        a = document.agent(EX_NS['ag4'])
        a.add_attributes([
            (PROV_LABEL, 'agent4'),
            (PROV_LABEL, Literal('hello')),
            (PROV_LABEL, Literal('bye', langtag="en")),
        ])
        self.do_tests(document)

    def test_agent_5(self):
        document = self.new_document()
        a = document.agent(EX_NS['ag5'])
        a.add_attributes([
            (PROV_LABEL, 'agent5'),
            (PROV_LABEL, Literal('hello')),
            (PROV_LABEL, Literal('bye', langtag="en")),
            (PROV_LABEL, Literal('bonjour', langtag="french")),
        ])
        self.do_tests(document)

    def test_agent_6(self):
        document = self.new_document()
        a = document.agent(EX_NS['ag6'])
        a.add_attributes([(PROV_LABEL, 'agent6')])
        self.add_types(a)
        self.do_tests(document)

    def test_agent_7(self):
        document = self.new_document()
        a = document.agent(EX_NS['ag7'])
        a.add_attributes([(PROV_LABEL, 'agent7')])
        self.add_locations(a)
        self.add_labels(a)
        self.do_tests(document)

    def test_agent_8(self):
        document = self.new_document()
        a = document.agent(EX_NS['ag8'])
        a.add_attributes([(PROV_LABEL, 'agent8')])
        self.add_types(a)
        self.add_locations(a)
        self.add_labels(a)
        self.add_further_attributes(a)
        self.do_tests(document)

    # GENERATIONS
    def test_generation_1(self):
        document = self.new_document()
        document.generation(EX_NS['e1'], identifier=EX_NS['gen1'])
        self.do_tests(document)

    def test_generation_2(self):
        document = self.new_document()
        document.generation(
            EX_NS['e1'], identifier=EX_NS['gen2'], activity=EX_NS['a1']
        )
        self.do_tests(document)

    def test_generation_3(self):
        document = self.new_document()
        a = document.generation(
            EX_NS['e1'], identifier=EX_NS['gen3'], activity=EX_NS['a1']
        )
        a.add_attributes([
            (PROV_ROLE, 'somerole'),
            (PROV_ROLE, 'otherRole'),
        ])
        self.do_tests(document)

    def test_generation_4(self):
        document = self.new_document()
        document.new_record(
            PROV_GENERATION, EX_NS['gen4'], (
                (PROV_ATTR_ENTITY, EX_NS['e1']),
                (PROV_ATTR_ACTIVITY, EX_NS['a1']),
                (PROV_ATTR_TIME, datetime.datetime.now())
            ),
            {PROV_ROLE: 'somerole'}
        )
        self.do_tests(document)

    def test_generation_5(self):
        document = self.new_document()
        a = document.generation(EX_NS['e1'],
                                identifier=EX_NS['gen5'],
                                activity=EX_NS['a1'],
                                time=datetime.datetime.now())
        a.add_attributes([
            (PROV_ROLE, 'somerole'),
        ])
        self.add_types(a)
        self.add_locations(a)
        self.add_labels(a)
        self.add_further_attributes(a)
        self.do_tests(document)

    def test_generation_6(self):
        document = self.new_document()
        document.generation(
            EX_NS['e1'], activity=EX_NS['a1'], time=datetime.datetime.now()
        )
        self.do_tests(document)

    def test_generation_7(self):
        document = self.new_document()
        a = document.generation(EX_NS['e1'],
                                activity=EX_NS['a1'],
                                time=datetime.datetime.now())
        a.add_attributes([(PROV_ROLE, 'somerole')])
        self.add_types(a)
        self.add_locations(a)
        self.add_labels(a)
        self.add_further_attributes(a)
        self.do_tests(document)

    # USAGE
    def test_usage_1(self):
        document = self.new_document()
        document.usage(
            None, entity=EX_NS['e1'], identifier=EX_NS['use1']
        )
        self.do_tests(document)

    def test_usage_2(self):
        document = self.new_document()
        document.usage(
            EX_NS['a1'], entity=EX_NS['e1'], identifier=EX_NS['use2']
        )
        self.do_tests(document)

    def test_usage_3(self):
        document = self.new_document()
        use = document.usage(
            EX_NS['a1'], entity=EX_NS['e1'], identifier=EX_NS['use3']
        )
        use.add_attributes([
            (PROV_ROLE, "somerole"),
            (PROV_ROLE, "otherRole")
        ])
        self.do_tests(document)

    def test_usage_4(self):
        document = self.new_document()
        use = document.usage(EX_NS['a1'],
                             entity=EX_NS['e1'],
                             identifier=EX_NS['use4'],
                             time=datetime.datetime.now())
        use.add_attributes([
            (PROV_ROLE, "somerole")
        ])
        self.do_tests(document)

    def test_usage_5(self):
        document = self.new_document()
        use = document.usage(EX_NS['a1'],
                             entity=EX_NS['e1'],
                             identifier=EX_NS['use5'],
                             time=datetime.datetime.now())
        use.add_attributes([
            (PROV_ROLE, "somerole")
        ])
        self.add_types(use)
        self.add_locations(use)
        self.add_labels(use)
        self.add_further_attributes(use)
        self.do_tests(document)

    def test_usage_6(self):
        document = self.new_document()
        document.usage(EX_NS['a1'], entity=EX_NS['e1'])
        self.do_tests(document)

    def test_usage_7(self):
        document = self.new_document()
        use = document.usage(EX_NS['a1'],
                             entity=EX_NS['e1'],
                             time=datetime.datetime.now())
        use.add_attributes([
            (PROV_ROLE, "somerole")
        ])
        self.add_types(use)
        self.add_locations(use)
        self.add_labels(use)
        self.add_further_attributes(use)
        self.do_tests(document)

    # INVALIDATIONS
    def test_invalidation_1(self):
        document = self.new_document()
        document.invalidation(EX_NS['e1'], identifier=EX_NS['inv1'])
        self.do_tests(document)

    def test_invalidation_2(self):
        document = self.new_document()
        document.invalidation(
            EX_NS['e1'], identifier=EX_NS['inv2'], activity=EX_NS['a1']
        )
        self.do_tests(document)

    def test_invalidation_3(self):
        document = self.new_document()
        inv = document.invalidation(
            EX_NS['e1'], identifier=EX_NS['inv3'], activity=EX_NS['a1']
        )
        inv.add_attributes([
            (PROV_ROLE, "someRole"),
            (PROV_ROLE, "otherRole"),
        ])
        self.do_tests(document)

    def test_invalidation_4(self):
        document = self.new_document()
        inv = document.invalidation(EX_NS['e1'],
                                    identifier=EX_NS['inv4'],
                                    activity=EX_NS['a1'],
                                    time=datetime.datetime.now())
        inv.add_attributes([
            (PROV_ROLE, "someRole"),
        ])
        self.do_tests(document)

    def test_invalidation_5(self):
        document = self.new_document()
        inv = document.invalidation(EX_NS['e1'],
                                    identifier=EX_NS['inv5'],
                                    activity=EX_NS['a1'],
                                    time=datetime.datetime.now())
        inv.add_attributes([
            (PROV_ROLE, "someRole"),
        ])
        self.add_types(inv)
        self.add_locations(inv)
        self.add_labels(inv)
        self.add_further_attributes(inv)
        self.do_tests(document)

    def test_invalidation_6(self):
        document = self.new_document()
        document.invalidation(EX_NS['e1'], activity=EX_NS['a1'])
        self.do_tests(document)

    def test_invalidation_7(self):
        document = self.new_document()
        inv = document.invalidation(EX_NS['e1'],
                                    activity=EX_NS['a1'],
                                    time=datetime.datetime.now())
        inv.add_attributes([
            (PROV_ROLE, "someRole"),
        ])
        self.add_types(inv)
        self.add_locations(inv)
        self.add_labels(inv)
        self.add_further_attributes(inv)
        self.do_tests(document)

    # STARTS
    def test_start_1(self):
        document = self.new_document()
        document.start(None, trigger=EX_NS['e1'], identifier=EX_NS['start1'])
        self.do_tests(document)

    def test_start_2(self):
        document = self.new_document()
        document.start(
            EX_NS['a1'], trigger=EX_NS['e1'], identifier=EX_NS['start2']
        )
        self.do_tests(document)

    def test_start_3(self):
        document = self.new_document()
        document.start(EX_NS['a1'], identifier=EX_NS['start3'])
        self.do_tests(document)

    def test_start_4(self):
        document = self.new_document()
        document.start(
            None, trigger=EX_NS['e1'], identifier=EX_NS['start4'],
            starter=EX_NS['a2']
        )
        self.do_tests(document)

    def test_start_5(self):
        document = self.new_document()
        document.start(
            EX_NS['a1'], trigger=EX_NS['e1'],
            identifier=EX_NS['start5'], starter=EX_NS['a2']
        )
        self.do_tests(document)

    def test_start_6(self):
        document = self.new_document()
        document.start(
            EX_NS['a1'], identifier=EX_NS['start6'], starter=EX_NS['a2']
        )
        self.do_tests(document)

    def test_start_7(self):
        document = self.new_document()
        document.start(
            EX_NS['a1'], identifier=EX_NS['start7'],
            starter=EX_NS['a2'], time=datetime.datetime.now()
        )
        self.do_tests(document)

    def test_start_8(self):
        document = self.new_document()
        start = document.start(EX_NS['a1'],
                               identifier=EX_NS['start8'],
                               starter=EX_NS['a2'],
                               time=datetime.datetime.now())
        start.add_attributes([
            (PROV_ROLE, "egg-cup"),
            (PROV_ROLE, "boiling-water"),
        ])
        self.add_types(start)
        self.add_locations(start)
        self.add_labels(start)
        self.add_further_attributes(start)
        self.do_tests(document)

    def test_start_9(self):
        document = self.new_document()
        document.start(EX_NS['a1'], trigger=EX_NS['e1'])
        self.do_tests(document)

    def test_start_10(self):
        document = self.new_document()
        start = document.start(EX_NS['a1'],
                               starter=EX_NS['a2'],
                               time=datetime.datetime.now())
        start.add_attributes([
            (PROV_ROLE, "egg-cup"),
            (PROV_ROLE, "boiling-water"),
        ])
        self.add_types(start)
        self.add_locations(start)
        self.add_labels(start)
        self.add_further_attributes(start)
        self.do_tests(document)

    # ENDS
    def test_end_1(self):
        document = self.new_document()
        document.end(None, trigger=EX_NS['e1'], identifier=EX_NS['end1'])
        self.do_tests(document)

    def test_end_2(self):
        document = self.new_document()
        document.end(
            EX_NS['a1'], trigger=EX_NS['e1'], identifier=EX_NS['end2']
        )
        self.do_tests(document)

    def test_end_3(self):
        document = self.new_document()
        document.end(EX_NS['a1'], identifier=EX_NS['end3'])
        self.do_tests(document)

    def test_end_4(self):
        document = self.new_document()
        document.end(
            None, trigger=EX_NS['e1'], identifier=EX_NS['end4'],
            ender=EX_NS['a2']
        )
        self.do_tests(document)

    def test_end_5(self):
        document = self.new_document()
        document.end(
            EX_NS['a1'], trigger=EX_NS['e1'], identifier=EX_NS['end5'],
            ender=EX_NS['a2']
        )
        self.do_tests(document)

    def test_end_6(self):
        document = self.new_document()
        document.end(EX_NS['a1'], identifier=EX_NS['end6'], ender=EX_NS['a2'])
        self.do_tests(document)

    def test_end_7(self):
        document = self.new_document()
        document.end(
            EX_NS['a1'], identifier=EX_NS['end7'], ender=EX_NS['a2'],
            time=datetime.datetime.now()
        )
        self.do_tests(document)

    def test_end_8(self):
        document = self.new_document()
        end = document.end(EX_NS['a1'],
                           identifier=EX_NS['end8'],
                           ender=EX_NS['a2'],
                           time=datetime.datetime.now())
        end.add_attributes([
            (PROV_ROLE, "egg-cup"),
            (PROV_ROLE, "boiling-water"),
        ])
        self.add_types(end)
        self.add_locations(end)
        self.add_labels(end)
        self.add_further_attributes(end)
        self.do_tests(document)

    def test_end_9(self):
        document = self.new_document()
        document.end(EX_NS['a1'], trigger=EX_NS['e1'])
        self.do_tests(document)

    def test_end_10(self):
        document = self.new_document()
        end = document.end(EX_NS['a1'],
                           ender=EX_NS['a2'],
                           time=datetime.datetime.now())
        end.add_attributes([
            (PROV_ROLE, "yolk"),
            (PROV_ROLE, "white"),
        ])
        self.add_types(end)
        self.add_locations(end)
        self.add_labels(end)
        self.add_further_attributes(end)
        self.do_tests(document)

    # DERIVATIONS
    def test_derivation_1(self):
        document = self.new_document()
        document.derivation(
            None, usedEntity=EX_NS['e1'], identifier=EX_NS['der1']
        )
        self.do_tests(document)

    def test_derivation_2(self):
        document = self.new_document()
        document.derivation(
            EX_NS['e2'], usedEntity=None, identifier=EX_NS['der2']
        )
        self.do_tests(document)

    def test_derivation_3(self):
        document = self.new_document()
        document.derivation(
            EX_NS['e2'], usedEntity=EX_NS['e1'], identifier=EX_NS['der3']
        )
        self.do_tests(document)

    def test_derivation_4(self):
        document = self.new_document()
        der = document.derivation(
            EX_NS['e2'], usedEntity=EX_NS['e1'], identifier=EX_NS['der4']
        )
        self.add_label(der)
        self.do_tests(document)

    def test_derivation_5(self):
        document = self.new_document()
        document.derivation(
            EX_NS['e2'], usedEntity=EX_NS['e1'],
            identifier=EX_NS['der5'], activity=EX_NS['a']
        )
        self.do_tests(document)

    def test_derivation_6(self):
        document = self.new_document()
        document.derivation(
            EX_NS['e2'], usedEntity=EX_NS['e1'],
            identifier=EX_NS['der6'], activity=EX_NS['a'],
            usage=EX_NS['u']
        )
        self.do_tests(document)

    def test_derivation_7(self):
        document = self.new_document()
        document.derivation(
            EX_NS['e2'], usedEntity=EX_NS['e1'],
            identifier=EX_NS['der7'], activity=EX_NS['a'],
            usage=EX_NS['u'], generation=EX_NS['g']
        )
        self.do_tests(document)

    def test_derivation_8(self):
        document = self.new_document()
        der = document.derivation(EX_NS['e2'],
                                  usedEntity=EX_NS['e1'],
                                  identifier=EX_NS['der8'])
        self.add_label(der)
        self.add_types(der)
        self.add_further_attributes(der)
        self.do_tests(document)

    def test_derivation_9(self):
        document = self.new_document()
        der = document.derivation(EX_NS['e2'], usedEntity=None)
        self.add_types(der)
        self.do_tests(document)

    def test_derivation_10(self):
        document = self.new_document()
        document.derivation(
            EX_NS['e2'], usedEntity=EX_NS['e1'],
            activity=EX_NS['a'], usage=EX_NS['u'],
            generation=EX_NS['g']
        )
        self.do_tests(document)

    def test_derivation_11(self):
        document = self.new_document()
        document.revision(
            EX_NS['e2'], usedEntity=EX_NS['e1'],
            identifier=EX_NS['rev1'], activity=EX_NS['a'],
            usage=EX_NS['u'], generation=EX_NS['g']
        )
        self.do_tests(document)

    def test_derivation_12(self):
        document = self.new_document()
        document.quotation(
            EX_NS['e2'], usedEntity=EX_NS['e1'],
            identifier=EX_NS['quo1'], activity=EX_NS['a'],
            usage=EX_NS['u'], generation=EX_NS['g']
        )
        self.do_tests(document)

    def test_derivation_13(self):
        document = self.new_document()
        document.primary_source(
            EX_NS['e2'], usedEntity=EX_NS['e1'],
            identifier=EX_NS['prim1'], activity=EX_NS['a'],
            usage=EX_NS['u'], generation=EX_NS['g']
        )
        self.do_tests(document)

    # ASSOCIATIONS
    def test_association_1(self):
        document = self.new_document()
        document.association(EX_NS['a1'], identifier=EX_NS['assoc1'])
        self.do_tests(document)

    def test_association_2(self):
        document = self.new_document()
        document.association(
            None, agent=EX_NS['ag1'], identifier=EX_NS['assoc2']
        )
        self.do_tests(document)

    def test_association_3(self):
        document = self.new_document()
        document.association(
            EX_NS['a1'], agent=EX_NS['ag1'], identifier=EX_NS['assoc3']
        )
        self.do_tests(document)

    def test_association_4(self):
        document = self.new_document()
        document.association(
            EX_NS['a1'], agent=EX_NS['ag1'],
            identifier=EX_NS['assoc4'], plan=EX_NS['plan1']
        )
        self.do_tests(document)

    def test_association_5(self):
        document = self.new_document()
        document.association(EX_NS['a1'], agent=EX_NS['ag1'])
        self.do_tests(document)

    def test_association_6(self):
        document = self.new_document()
        assoc = document.association(EX_NS['a1'],
                                     agent=EX_NS['ag1'],
                                     identifier=EX_NS['assoc6'],
                                     plan=EX_NS['plan1'])
        self.add_labels(assoc)
        self.do_tests(document)

    def test_association_7(self):
        document = self.new_document()
        assoc = document.association(EX_NS['a1'],
                                     agent=EX_NS['ag1'],
                                     identifier=EX_NS['assoc7'],
                                     plan=EX_NS['plan1'])
        self.add_labels(assoc)
        self.add_types(assoc)
        self.do_tests(document)

    def test_association_8(self):
        document = self.new_document()
        assoc = document.association(EX_NS['a1'],
                                     agent=EX_NS['ag1'],
                                     identifier=EX_NS['assoc8'],
                                     plan=EX_NS['plan1'])
        assoc.add_attributes([
            (PROV_ROLE, "figroll"),
            (PROV_ROLE, "sausageroll"),
        ])
        self.do_tests(document)

    def test_association_9(self):
        document = self.new_document()
        assoc = document.association(EX_NS['a1'],
                                     agent=EX_NS['ag1'],
                                     identifier=EX_NS['assoc9'],
                                     plan=EX_NS['plan1'])
        self.add_labels(assoc)
        self.add_types(assoc)
        self.add_further_attributes(assoc)
        self.do_tests(document)

    def test_association_10(self):
        document = self.new_document()
        assoc1 = document.association(EX_NS['a1'],
                                      agent=EX_NS['ag1'],
                                      identifier=EX_NS['assoc10a'])
        assoc1.add_attributes([
            (PROV_ROLE, "figroll"),
        ])
        assoc2 = document.association(EX_NS['a1'],
                                      agent=EX_NS['ag2'],
                                      identifier=EX_NS['assoc10b'])
        assoc2.add_attributes([
            (PROV_ROLE, "sausageroll"),
        ])
        self.do_tests(document)

    # ATTRIBUTIONS
    def test_attribution_1(self):
        document = self.new_document()
        document.attribution(
            EX_NS['e1'], None, identifier=EX_NS['attr1']
        )
        self.do_tests(document)

    def test_attribution_2(self):
        document = self.new_document()
        document.attribution(
            None, EX_NS['ag1'], identifier=EX_NS['attr2']
        )
        self.do_tests(document)

    def test_attribution_3(self):
        document = self.new_document()
        document.attribution(
            EX_NS['e1'], EX_NS['ag1'], identifier=EX_NS['attr3']
        )
        self.do_tests(document)

    def test_attribution_4(self):
        document = self.new_document()
        document.attribution(
            EX_NS['e1'], EX_NS['ag1'], identifier=EX_NS['attr4']
        )
        self.do_tests(document)

    def test_attribution_5(self):
        document = self.new_document()
        document.attribution(EX_NS['e1'], EX_NS['ag1'])
        self.do_tests(document)

    def test_attribution_6(self):
        document = self.new_document()
        attr = document.attribution(
            EX_NS['e1'], EX_NS['ag1'], identifier=EX_NS['attr6']
        )
        self.add_labels(attr)
        self.do_tests(document)

    def test_attribution_7(self):
        document = self.new_document()
        attr = document.attribution(
            EX_NS['e1'], EX_NS['ag1'], identifier=EX_NS['attr7']
        )
        self.add_labels(attr)
        self.add_types(attr)
        self.do_tests(document)

    def test_attribution_8(self):
        document = self.new_document()
        attr = document.attribution(
            EX_NS['e1'], EX_NS['ag1'], identifier=EX_NS['attr8']
        )
        self.add_labels(attr)
        self.add_types(attr)
        self.add_further_attributes(attr)
        self.do_tests(document)

    # DELEGATIONS
    def test_delegation_1(self):
        document = self.new_document()
        document.delegation(
            EX_NS['e1'], None, identifier=EX_NS['dele1']
        )
        self.do_tests(document)

    def test_delegation_2(self):
        document = self.new_document()
        document.delegation(
            None, EX_NS['ag1'], identifier=EX_NS['dele2']
        )
        self.do_tests(document)

    def test_delegation_3(self):
        document = self.new_document()
        document.delegation(
            EX_NS['e1'], EX_NS['ag1'], identifier=EX_NS['dele3']
        )
        self.do_tests(document)

    def test_delegation_4(self):
        document = self.new_document()
        document.delegation(
            EX_NS['e1'], EX_NS['ag1'],
            activity=EX_NS['a1'], identifier=EX_NS['dele4']
        )
        self.do_tests(document)

    def test_delegation_5(self):
        document = self.new_document()
        document.delegation(EX_NS['e1'], EX_NS['ag1'])
        self.do_tests(document)

    def test_delegation_6(self):
        document = self.new_document()
        dele = document.delegation(EX_NS['e1'],
                                   EX_NS['ag1'],
                                   activity=EX_NS['a1'],
                                   identifier=EX_NS['dele6'])
        self.add_labels(dele)
        self.do_tests(document)

    def test_delegation_7(self):
        document = self.new_document()
        dele = document.delegation(EX_NS['e1'],
                                   EX_NS['ag1'],
                                   activity=EX_NS['a1'],
                                   identifier=EX_NS['dele7'])
        self.add_labels(dele)
        self.add_types(dele)
        self.do_tests(document)

    def test_delegation_8(self):
        document = self.new_document()
        dele = document.delegation(EX_NS['e1'],
                                   EX_NS['ag1'],
                                   activity=EX_NS['a1'],
                                   identifier=EX_NS['dele8'])
        self.add_labels(dele)
        self.add_types(dele)
        self.add_further_attributes(dele)
        self.do_tests(document)

    # COMMUNICATIONS
    def test_communication_1(self):
        document = self.new_document()
        document.communication(
            EX_NS['a2'], None, identifier=EX_NS['inf1']
        )
        self.do_tests(document)

    def test_communication_2(self):
        document = self.new_document()
        document.communication(None, EX_NS['a1'], identifier=EX_NS['inf2'])
        self.do_tests(document)

    def test_communication_3(self):
        document = self.new_document()
        document.communication(
            EX_NS['a2'], EX_NS['a1'], identifier=EX_NS['inf3']
        )
        self.do_tests(document)

    def test_communication_4(self):
        document = self.new_document()
        document.communication(EX_NS['a2'], EX_NS['a1'])
        self.do_tests(document)

    def test_communication_5(self):
        document = self.new_document()
        inf = document.communication(
            EX_NS['a2'], EX_NS['a1'], identifier=EX_NS['inf5']
        )
        self.add_labels(inf)
        self.do_tests(document)

    def test_communication_6(self):
        document = self.new_document()
        inf = document.communication(EX_NS['a2'], EX_NS['a1'],
                                     identifier=EX_NS['inf6'])
        self.add_labels(inf)
        self.add_types(inf)
        self.do_tests(document)

    def test_communication_7(self):
        document = self.new_document()
        inf = document.communication(EX_NS['a2'], EX_NS['a1'],
                                     identifier=EX_NS['inf7'])
        self.add_labels(inf)
        self.add_types(inf)
        self.add_further_attributes(inf)
        self.do_tests(document)

    # INFLUENCES
    def test_influence_1(self):
        document = self.new_document()
        document.influence(EX_NS['a2'], None, identifier=EX_NS['inf1'])
        self.do_tests(document)

    def test_influence_2(self):
        document = self.new_document()
        document.influence(None, EX_NS['a1'], identifier=EX_NS['inf2'])
        self.do_tests(document)

    def test_influence_3(self):
        document = self.new_document()
        document.influence(EX_NS['a2'], EX_NS['a1'], identifier=EX_NS['inf3'])
        self.do_tests(document)

    def test_influence_4(self):
        document = self.new_document()
        document.influence(EX_NS['a2'], EX_NS['a1'])
        self.do_tests(document)

    def test_influence_5(self):
        document = self.new_document()
        inf = document.influence(EX_NS['a2'], EX_NS['a1'],
                                 identifier=EX_NS['inf5'])
        self.add_labels(inf)
        self.do_tests(document)

    def test_influence_6(self):
        document = self.new_document()
        inf = document.influence(EX_NS['a2'], EX_NS['a1'],
                                 identifier=EX_NS['inf6'])
        self.add_labels(inf)
        self.add_types(inf)
        self.do_tests(document)

    def test_influence_7(self):
        document = self.new_document()
        inf = document.influence(EX_NS['a2'], EX_NS['a1'],
                                 identifier=EX_NS['inf7'])
        self.add_labels(inf)
        self.add_types(inf)
        self.add_further_attributes(inf)
        self.do_tests(document)

    # OTHERS
    def test_alternate_1(self):
        document = self.new_document()
        document.alternate(EX_NS['e2'], EX_NS['e1'])
        self.do_tests(document)

    def test_specialization_1(self):
        document = self.new_document()
        document.specialization(EX_NS['e2'], EX_NS['e1'])
        self.do_tests(document)

    def test_mention_1(self):
        document = self.new_document()
        document.mention(EX_NS['e2'], EX_NS['e1'], None)
        self.do_tests(document)

    def test_mention_2(self):
        document = self.new_document()
        document.mention(EX_NS['e2'], EX_NS['e1'], EX_NS['b'])
        self.do_tests(document)

    def test_membership_1(self):
        document = self.new_document()
        document.membership(EX_NS['c'], EX_NS['e1'])
        self.do_tests(document)

    def test_membership_2(self):
        document = self.new_document()
        document.membership(EX_NS['c'], EX_NS['e1'])
        document.membership(EX_NS['c'], EX_NS['e2'])
        self.do_tests(document)

    def test_membership_3(self):
        document = self.new_document()
        document.membership(EX_NS['c'], EX_NS['e1'])
        document.membership(EX_NS['c'], EX_NS['e2'])
        document.membership(EX_NS['c'], EX_NS['e3'])
        self.do_tests(document)

    # SCRUFFY
    def test_scruffy_generation_1(self):
        document = self.new_document()
        document.generation(
            EX_NS['e1'], EX_NS['a1'],
            identifier=EX_NS['gen1'],
            time=datetime.datetime.now()
        )
        document.generation(
            EX_NS['e1'], EX_NS['a1'],
            identifier=EX_NS['gen1'],
            time=dateutil.parser.parse("2012-12-03T21:08:16.686Z")
        )
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_generation_2(self):
        document = self.new_document()
        gen1 = document.generation(EX_NS['e1'],
                                   EX_NS['a1'],
                                   identifier=EX_NS['gen1'],
                                   time=datetime.datetime.now())
        gen2 = document.generation(EX_NS['e1'],
                                   EX_NS['a1'],
                                   identifier=EX_NS['gen1'],
                                   time=dateutil.parser.parse(
                                       "2012-12-03T21:08:16.686Z"))
        gen1.add_attributes([
            (EX_NS['tag2'], "hello-scruff-gen2")
        ])
        gen2.add_attributes([
            (EX_NS['tag2'], "hi-scruff-gen2")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_invalidation_1(self):
        document = self.new_document()
        document.invalidation(
            EX_NS['e1'], EX_NS['a1'],
            identifier=EX_NS['gen1'],
            time=datetime.datetime.now()
        )
        document.invalidation(
            EX_NS['e1'], EX_NS['a1'],
            identifier=EX_NS['gen1'],
            time=dateutil.parser.parse("2012-12-03T21:08:16.686Z")
        )
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_invalidation_2(self):
        document = self.new_document()
        inv1 = document.invalidation(EX_NS['e1'],
                                     EX_NS['a1'],
                                     identifier=EX_NS['gen1'],
                                     time=datetime.datetime.now())
        inv2 = document.invalidation(EX_NS['e1'],
                                     EX_NS['a1'],
                                     identifier=EX_NS['gen1'],
                                     time=dateutil.parser.parse(
                                         "2012-12-03T21:08:16.686Z"))
        inv1.add_attributes([
            (EX_NS['tag2'], "hello")
        ])
        inv2.add_attributes([
            (EX_NS['tag2'], "hi")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_usage_1(self):
        document = self.new_document()
        document.usage(
            EX_NS['a1'], EX_NS['e1'],
            identifier=EX_NS['gen1'],
            time=datetime.datetime.now()
        )
        document.usage(
            EX_NS['a1'], EX_NS['e1'],
            identifier=EX_NS['gen1'],
            time=dateutil.parser.parse("2012-12-03T21:08:16.686Z")
        )
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_usage_2(self):
        document = self.new_document()
        use1 = document.usage(EX_NS['a1'],
                              EX_NS['e1'],
                              identifier=EX_NS['gen1'],
                              time=datetime.datetime.now())
        use2 = document.usage(EX_NS['a1'],
                              EX_NS['e1'],
                              identifier=EX_NS['gen1'],
                              time=dateutil.parser.parse(
                                  "2012-12-03T21:08:16.686Z"))
        use1.add_attributes([
            (EX_NS['tag2'], "hello")
        ])
        use2.add_attributes([
            (EX_NS['tag2'], "hi")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_start_1(self):
        document = self.new_document()
        document.start(
            EX_NS['a1'], EX_NS['e1'],
            identifier=EX_NS['gen1'],
            time=datetime.datetime.now()
        )
        document.start(
            EX_NS['a1'], EX_NS['e1'],
            identifier=EX_NS['gen1'],
            time=dateutil.parser.parse("2012-12-03T21:08:16.686Z")
        )
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_start_2(self):
        document = self.new_document()
        start1 = document.start(EX_NS['a1'],
                                EX_NS['e1'],
                                identifier=EX_NS['gen1'],
                                time=datetime.datetime.now())
        start2 = document.start(EX_NS['a1'],
                                EX_NS['e1'],
                                identifier=EX_NS['gen1'],
                                time=dateutil.parser.parse(
                                    "2012-12-03T21:08:16.686Z"))
        start1.add_attributes([
            (EX_NS['tag2'], "hello")
        ])
        start2.add_attributes([
            (EX_NS['tag2'], "hi")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_start_3(self):
        document = self.new_document()
        start1 = document.start(EX_NS['a1'],
                                EX_NS['e1'],
                                identifier=EX_NS['gen1'],
                                time=datetime.datetime.now(),
                                starter=EX_NS['a1s'])
        start2 = document.start(EX_NS['a1'],
                                EX_NS['e1'],
                                identifier=EX_NS['gen1'],
                                time=dateutil.parser.parse(
                                    "2012-12-03T21:08:16.686Z"),
                                starter=EX_NS['a2s'])
        start1.add_attributes([
            (EX_NS['tag2'], "hello")
        ])
        start2.add_attributes([
            (EX_NS['tag2'], "hi")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        document.activity(identifier=EX_NS['a2'])
        document.activity(identifier=EX_NS['a2s'])
        self.do_tests(document)

    def test_scruffy_start_4(self):
        document = self.new_document()
        start1 = document.start(EX_NS['a1'],
                                EX_NS['e1'],
                                identifier=EX_NS['gen1'],
                                time=datetime.datetime.now(),
                                starter=EX_NS['a1s'])
        start2 = document.start(EX_NS['a1'],
                                EX_NS['e1'],
                                identifier=EX_NS['gen1'],
                                time=dateutil.parser.parse(
                                    "2012-12-03T21:08:16.686Z"),
                                starter=EX_NS['a2s'])
        start1.add_attributes([
            (EX_NS['tag2'], "hello")
        ])
        start2.add_attributes([
            (EX_NS['tag2'], "hi")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        document.activity(identifier=EX_NS['a1s'])
        document.activity(identifier=EX_NS['a2'])
        document.activity(identifier=EX_NS['a2s'])
        self.do_tests(document)

    def test_scruffy_end_1(self):
        document = self.new_document()
        document.end(
            EX_NS['a1'],
            EX_NS['e1'],
            identifier=EX_NS['gen1'],
            time=datetime.datetime.now()
        )
        document.end(
            EX_NS['a1'],
            EX_NS['e1'],
            identifier=EX_NS['gen1'],
            time=dateutil.parser.parse("2012-12-03T21:08:16.686Z")
        )
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_end_2(self):
        document = self.new_document()
        end1 = document.end(EX_NS['a1'],
                            EX_NS['e1'],
                            identifier=EX_NS['gen1'],
                            time=datetime.datetime.now())
        end2 = document.end(EX_NS['a1'],
                            EX_NS['e1'],
                            identifier=EX_NS['gen1'],
                            time=dateutil.parser.parse(
                                "2012-12-03T21:08:16.686Z"))
        end1.add_attributes([
            (EX_NS['tag2'], "hello")
        ])
        end2.add_attributes([
            (EX_NS['tag2'], "hi")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        self.do_tests(document)

    def test_scruffy_end_3(self):
        document = self.new_document()
        end1 = document.end(EX_NS['a1'],
                            EX_NS['e1'],
                            identifier=EX_NS['gen1'],
                            time=datetime.datetime.now(),
                            ender=EX_NS['a1s'])
        end2 = document.end(EX_NS['a1'],
                            EX_NS['e1'],
                            identifier=EX_NS['gen1'],
                            time=dateutil.parser.parse(
                                "2012-12-03T21:08:16.686Z"),
                            ender=EX_NS['a2s'])
        end1.add_attributes([
            (EX_NS['tag2'], "hello")
        ])
        end2.add_attributes([
            (EX_NS['tag2'], "hi")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        document.activity(identifier=EX_NS['a2'])
        document.activity(identifier=EX_NS['a2s'])
        self.do_tests(document)

    def test_scruffy_end_4(self):
        document = self.new_document()
        end1 = document.end(EX_NS['a1'],
                            EX_NS['e1'],
                            identifier=EX_NS['gen1'],
                            time=datetime.datetime.now(),
                            ender=EX_NS['a1s'])
        end2 = document.end(EX_NS['a1'],
                            EX_NS['e1'],
                            identifier=EX_NS['gen1'],
                            time=dateutil.parser.parse(
                                "2012-12-03T21:08:16.686Z"),
                            ender=EX_NS['a2s'])
        end1.add_attributes([
            (EX_NS['tag2'], "hello")
        ])
        end2.add_attributes([
            (EX_NS['tag2'], "hi")
        ])
        document.entity(identifier=EX_NS['e1'])
        document.activity(identifier=EX_NS['a1'])
        document.activity(identifier=EX_NS['a1s'])
        document.activity(identifier=EX_NS['a2'])
        document.activity(identifier=EX_NS['a2s'])
        self.do_tests(document)

    def test_bundle_1(self):
        document = self.new_document()

        bundle1 = ProvBundle(identifier=EX_NS['bundle1'])
        bundle1.usage(activity=EX_NS['a1'], entity=EX_NS['e1'],
                      identifier=EX_NS['use1'])
        bundle1.entity(identifier=EX_NS['e1'])
        bundle1.activity(identifier=EX_NS['a1'])

        bundle2 = ProvBundle(identifier=EX_NS['bundle2'])
        bundle2.usage(activity=EX_NS['aa1'], entity=EX_NS['ee1'],
                      identifier=EX_NS['use2'])
        bundle2.entity(identifier=EX_NS['ee1'])
        bundle2.activity(identifier=EX_NS['aa1'])

        document.add_bundle(bundle1)
        document.add_bundle(bundle2)

        self.do_tests(document)

    def test_bundle_2(self):
        document = self.new_document()

        bundle1 = ProvBundle(identifier=EX_NS['bundle1'])
        bundle1.usage(activity=EX_NS['a1'], entity=EX_NS['e1'],
                      identifier=EX_NS['use1'])
        bundle1.entity(identifier=EX_NS['e1'])
        bundle1.activity(identifier=EX_NS['a1'])

        bundle2 = ProvBundle(identifier=EX_NS['bundle2'])
        bundle2.usage(activity=EX_NS['a1'], entity=EX_NS['e1'],
                      identifier=EX_NS['use2'])
        bundle2.entity(identifier=EX_NS['e1'])
        bundle2.activity(identifier=EX_NS['a1'])

        document.add_bundle(bundle1)
        document.add_bundle(bundle2)

        self.do_tests(document)

    def test_bundle_3(self):
        document = self.new_document()

        bundle1 = ProvBundle(identifier=EX_NS['bundle1'])
        bundle1.usage(activity=EX_NS['a1'], entity=EX_NS['e1'],
                      identifier=EX_NS['use1'])
        bundle1.entity(identifier=EX_NS['e1'])
        bundle1.activity(identifier=EX_NS['a1'])

        bundle2 = ProvBundle(identifier=EX_NS['bundle2'])
        bundle2.usage(activity=EX_NS['aa1'], entity=EX_NS['ee1'],
                      identifier=EX_NS['use2'])
        bundle2.entity(identifier=EX_NS['ee1'])
        bundle2.activity(identifier=EX_NS['aa1'])

        document.add_bundle(bundle1)
        document.add_bundle(bundle2)

        self.do_tests(document)

    def test_bundle_4(self):
        document = self.new_document()

        bundle1 = ProvBundle(identifier=EX_NS['bundle1'])
        bundle1.usage(activity=EX_NS['a1'], entity=EX_NS['e1'],
                      identifier=EX_NS['use1'])
        bundle1.entity(identifier=EX_NS['e1'])
        bundle1.activity(identifier=EX_NS['a1'])

        bundle2 = ProvBundle(identifier=EX_NS['bundle2'])
        bundle2.usage(activity=EX2_NS['aa1'], entity=EX2_NS['ee1'],
                      identifier=EX2_NS['use2'])
        bundle2.entity(identifier=EX2_NS['ee1'])
        bundle2.activity(identifier=EX2_NS['aa1'])

        document.add_bundle(bundle1)
        document.add_bundle(bundle2)

        self.do_tests(document)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from prov.model import *


EX_NS = Namespace('ex', 'http://example.org/')
EX_OTHER_NS = Namespace('other', 'http://example.org/')


class TestAttributesBase(object):
    """This is the base class for testing support for various datatypes.
    It is not runnable and needs to be included in a subclass of
    RoundTripTestCase.
    """

    attribute_values = [
        "un lieu",
        Literal("un lieu", langtag='fr'),
        Literal("a place", langtag='en'),
        Literal(1, XSD_INT),
        Literal(1, XSD_LONG),
        Literal(1, XSD_SHORT),
        Literal(2.0, XSD_DOUBLE),
        Literal(1.0, XSD_FLOAT),
        Literal(10, XSD_DECIMAL),
        True,
        False,
        Literal(10, XSD_BYTE),
        Literal(10, XSD_UNSIGNEDINT),
        Literal(10, XSD_UNSIGNEDLONG),
        Literal(10, XSD_INTEGER),
        Literal(10, XSD_UNSIGNEDSHORT),
        Literal(10, XSD_NONNEGATIVEINTEGER),
        Literal(-10, XSD_NONPOSITIVEINTEGER),
        Literal(10, XSD_POSITIVEINTEGER),
        Literal(10, XSD_UNSIGNEDBYTE),
        Identifier('http://example.org'),
        Literal('http://example.org', XSD_ANYURI),
        EX_NS['abc'],
        EX_OTHER_NS['abcd'],
        Namespace('ex', 'http://example4.org/')['zabc'],
        Namespace('other', 'http://example4.org/')['zabcd'],

        datetime.datetime.now(),
        Literal(datetime.datetime.now().isoformat(), XSD_DATETIME)
    ]

    def new_document(self):
        return ProvDocument()

    def run_entity_with_one_type_attribute(self, n):
        document = self.new_document()
        document.entity(
            EX_NS['et%d' % n], {'prov:type': self.attribute_values[n]}
        )
        self.do_tests(document)

    def test_entity_with_one_type_attribute_0(self):
        self.run_entity_with_one_type_attribute(0)

    def test_entity_with_one_type_attribute_1(self):
        self.run_entity_with_one_type_attribute(1)

    def test_entity_with_one_type_attribute_2(self):
        self.run_entity_with_one_type_attribute(2)

    def test_entity_with_one_type_attribute_3(self):
        self.run_entity_with_one_type_attribute(3)

    def test_entity_with_one_type_attribute_4(self):
        self.run_entity_with_one_type_attribute(4)

    def test_entity_with_one_type_attribute_5(self):
        self.run_entity_with_one_type_attribute(5)

    def test_entity_with_one_type_attribute_6(self):
        self.run_entity_with_one_type_attribute(6)

    def test_entity_with_one_type_attribute_7(self):
        self.run_entity_with_one_type_attribute(7)

    def test_entity_with_one_type_attribute_8(self):
        self.run_entity_with_one_type_attribute(8)

    def test_entity_with_one_type_attribute_9(self):
        self.run_entity_with_one_type_attribute(9)

    def test_entity_with_one_type_attribute_10(self):
        self.run_entity_with_one_type_attribute(10)

    def test_entity_with_one_type_attribute_11(self):
        self.run_entity_with_one_type_attribute(11)

    def test_entity_with_one_type_attribute_12(self):
        self.run_entity_with_one_type_attribute(12)

    def test_entity_with_one_type_attribute_13(self):
        self.run_entity_with_one_type_attribute(13)

    def test_entity_with_one_type_attribute_14(self):
        self.run_entity_with_one_type_attribute(14)

    def test_entity_with_one_type_attribute_15(self):
        self.run_entity_with_one_type_attribute(15)

    def test_entity_with_one_type_attribute_16(self):
        self.run_entity_with_one_type_attribute(16)

    def test_entity_with_one_type_attribute_17(self):
        self.run_entity_with_one_type_attribute(17)

    def test_entity_with_one_type_attribute_18(self):
        self.run_entity_with_one_type_attribute(18)

    def test_entity_with_one_type_attribute_19(self):
        self.run_entity_with_one_type_attribute(19)

    def test_entity_with_one_type_attribute_20(self):
        self.run_entity_with_one_type_attribute(20)

    def test_entity_with_one_type_attribute_21(self):
        self.run_entity_with_one_type_attribute(21)

    def test_entity_with_one_type_attribute_22(self):
        self.run_entity_with_one_type_attribute(22)

    def test_entity_with_one_type_attribute_23(self):
        self.run_entity_with_one_type_attribute(23)

    def test_entity_with_one_type_attribute_24(self):
        self.run_entity_with_one_type_attribute(24)

    def test_entity_with_one_type_attribute_25(self):
        self.run_entity_with_one_type_attribute(25)

    def test_entity_with_one_type_attribute_26(self):
        self.run_entity_with_one_type_attribute(26)

    def test_entity_with_one_type_attribute_27(self):
        self.run_entity_with_one_type_attribute(27)

    def test_entity_with_multiple_attribute(self):
        document = self.new_document()
        attributes = [
            (EX_NS['v_%d' % i], value)
            for i, value in enumerate(self.attribute_values)
        ]
        document.entity(EX_NS['emov'], attributes)
        self.do_tests(document)

    def test_entity_with_multiple_value_attribute(self):
        document = self.new_document()
        attributes = [
            ('prov:value', value)
            for i, value in enumerate(self.attribute_values)
        ]
        document.entity(EX_NS['emv'], attributes)
        self.do_tests(document)



# coding: utf8
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from prov.model import ProvDocument, Namespace, Literal, PROV, Identifier
import datetime


def primer_example():
    # https://github.com/lucmoreau/ProvToolbox/blob/master/prov-n/src/test/resources/prov/primer.pn
    #===========================================================================
    # document
    g = ProvDocument()

    #    prefix ex <http://example/>
    #    prefix dcterms <http://purl.org/dc/terms/>
    #    prefix foaf <http://xmlns.com/foaf/0.1/>
    ex = Namespace('ex', 'http://example/')  # namespaces do not need to be explicitly added to a document
    g.add_namespace("dcterms", "http://purl.org/dc/terms/")
    g.add_namespace("foaf", "http://xmlns.com/foaf/0.1/")

    #    entity(ex:article, [dcterms:title="Crime rises in cities"])
    # first time the ex namespace was used, it is added to the document automatically
    g.entity(ex['article'], {'dcterms:title': "Crime rises in cities"})
    #    entity(ex:articleV1)
    g.entity(ex['articleV1'])
    #    entity(ex:articleV2)
    g.entity(ex['articleV2'])
    #    entity(ex:dataSet1)
    g.entity(ex['dataSet1'])
    #    entity(ex:dataSet2)
    g.entity(ex['dataSet2'])
    #    entity(ex:regionList)
    g.entity(ex['regionList'])
    #    entity(ex:composition)
    g.entity(ex['composition'])
    #    entity(ex:chart1)
    g.entity(ex['chart1'])
    #    entity(ex:chart2)
    g.entity(ex['chart2'])
    #    entity(ex:blogEntry)
    g.entity(ex['blogEntry'])

    #    activity(ex:compile)
    g.activity('ex:compile')  # since ex is registered, it can be used like this
    #    activity(ex:compile2)
    g.activity('ex:compile2')
    #    activity(ex:compose)
    g.activity('ex:compose')
    #    activity(ex:correct, 2012-03-31T09:21:00, 2012-04-01T15:21:00)
    g.activity('ex:correct', '2012-03-31T09:21:00', '2012-04-01T15:21:00')  # date time can be provided as strings
    #    activity(ex:illustrate)
    g.activity('ex:illustrate')

    #    used(ex:compose, ex:dataSet1, -,   [ prov:role = "ex:dataToCompose"])
    g.used('ex:compose', 'ex:dataSet1', other_attributes={'prov:role': "ex:dataToCompose"})
    #    used(ex:compose, ex:regionList, -, [ prov:role = "ex:regionsToAggregateBy"])
    g.used('ex:compose', 'ex:regionList', other_attributes={'prov:role': "ex:regionsToAggregateBy"})
    #    wasGeneratedBy(ex:composition, ex:compose, -)
    g.wasGeneratedBy('ex:composition', 'ex:compose')

    #    used(ex:illustrate, ex:composition, -)
    g.used('ex:illustrate', 'ex:composition')
    #    wasGeneratedBy(ex:chart1, ex:illustrate, -)
    g.wasGeneratedBy('ex:chart1', 'ex:illustrate')

    #    wasGeneratedBy(ex:chart1, ex:compile,  2012-03-02T10:30:00)
    g.wasGeneratedBy('ex:chart1', 'ex:compile', '2012-03-02T10:30:00')
    #    wasGeneratedBy(ex:chart2, ex:compile2, 2012-04-01T15:21:00)
    #
    #
    #    agent(ex:derek, [ prov:type="prov:Person", foaf:givenName = "Derek",
    #           foaf:mbox= "<mailto:derek@example.org>"])
    g.agent('ex:derek', {
        'prov:type': PROV["Person"], 'foaf:givenName': "Derek", 'foaf:mbox': "<mailto:derek@example.org>"
    })
    #    wasAssociatedWith(ex:compose, ex:derek, -)
    g.wasAssociatedWith('ex:compose', 'ex:derek')
    #    wasAssociatedWith(ex:illustrate, ex:derek, -)
    g.wasAssociatedWith('ex:illustrate', 'ex:derek')
    #
    #    agent(ex:chartgen, [ prov:type="prov:Organization",
    #           foaf:name = "Chart Generators Inc"])
    g.agent('ex:chartgen', {'prov:type': PROV["Organization"], 'foaf:name': "Chart Generators Inc"})
    #    actedOnBehalfOf(ex:derek, ex:chartgen, ex:compose)
    g.actedOnBehalfOf('ex:derek', 'ex:chartgen', 'ex:compose')
    #    wasAttributedTo(ex:chart1, ex:derek)
    g.wasAttributedTo('ex:chart1', 'ex:derek')

    #    wasGeneratedBy(ex:dataSet2, ex:correct, -)
    g.wasGeneratedBy('ex:dataSet2', 'ex:correct')
    #    used(ex:correct, ex:dataSet1, -)
    g.used('ex:correct', 'ex:dataSet1')
    #    wasDerivedFrom(ex:dataSet2, ex:dataSet1, [prov:type='prov:Revision'])
    g.wasDerivedFrom('ex:dataSet2', 'ex:dataSet1', other_attributes={'prov:type': PROV['Revision']})
    #    wasDerivedFrom(ex:chart2, ex:dataSet2)
    g.wasDerivedFrom('ex:chart2', 'ex:dataSet2')

    #    wasDerivedFrom(ex:blogEntry, ex:article, [prov:type='prov:Quotation'])
    g.wasDerivedFrom('ex:blogEntry', 'ex:article', other_attributes={'prov:type': PROV['Quotation']})
    #    specializationOf(ex:articleV1, ex:article)
    g.specializationOf('ex:articleV1', 'ex:article')
    #    wasDerivedFrom(ex:articleV1, ex:dataSet1)
    g.wasDerivedFrom('ex:articleV1', 'ex:dataSet1')

    #    specializationOf(ex:articleV2, ex:article)
    g.specializationOf('ex:articleV2', 'ex:article')
    #    wasDerivedFrom(ex:articleV2, ex:dataSet2)
    g.wasDerivedFrom('ex:articleV2', 'ex:dataSet2')

    #    alternateOf(ex:articleV2, ex:articleV1)
    g.alternateOf('ex:articleV2', 'ex:articleV1')

    # endDocument
    return g


def primer_example_alternate():
    g = ProvDocument(namespaces={
        'ex': 'http://example/',
        'dcterms': 'http://purl.org/dc/terms/',
        'foaf': 'http://xmlns.com/foaf/0.1/'
    })

    article = g.entity('ex:article', {'dcterms:title': "Crime rises in cities"})
    articleV1 = g.entity('ex:articleV1')
    articleV2 = g.entity('ex:articleV2')
    dataSet1 = g.entity('ex:dataSet1')
    dataSet2 = g.entity('ex:dataSet2')
    regionList = g.entity('ex:regionList')
    composition = g.entity('ex:composition')
    chart1 = g.entity('ex:chart1')
    chart2 = g.entity('ex:chart2')
    blogEntry = g.entity('ex:blogEntry')

    compile = g.activity('ex:compile')
    compile2 = g.activity('ex:compile2')
    compose = g.activity('ex:compose')
    correct = g.activity('ex:correct', '2012-03-31T09:21:00', '2012-04-01T15:21:00')
    illustrate = g.activity('ex:illustrate')

    compose.used(dataSet1, attributes={'prov:role': "ex:dataToCompose"})
    compose.used(regionList, attributes={'prov:role': "ex:regionsToAggregateBy"})
    composition.wasGeneratedBy(compose)

    illustrate.used(composition)
    chart1.wasGeneratedBy(illustrate)

    chart1.wasGeneratedBy(compile, '2012-03-02T10:30:00')

    derek = g.agent('ex:derek', {
        'prov:type': PROV['Person'], 'foaf:givenName': "Derek", 'foaf:mbox': "<mailto:derek@example.org>"
    })
    compose.wasAssociatedWith(derek)
    illustrate.wasAssociatedWith(derek)

    chartgen = g.agent('ex:chartgen', {
        'prov:type': PROV["Organization"], 'foaf:name': "Chart Generators Inc"
    })
    derek.actedOnBehalfOf(chartgen, compose)
    chart1.wasAttributedTo(derek)

    dataSet2.wasGeneratedBy(correct)
    correct.used(dataSet1)
    dataSet2.wasDerivedFrom(dataSet1, attributes={'prov:type': PROV['Revision']})
    chart2.wasDerivedFrom(dataSet2)

    blogEntry.wasDerivedFrom(article, attributes={'prov:type': PROV['Quotation']})
    articleV1.specializationOf(article)
    articleV1.wasDerivedFrom(dataSet1)

    articleV2.specializationOf(article)
    articleV2.wasDerivedFrom(dataSet2)

    articleV2.alternateOf(articleV1)

    return g


def w3c_publication_1():
    # https://github.com/lucmoreau/ProvToolbox/blob/master/asn/src/test/resources/prov/w3c-publication1.prov-asn
    #===========================================================================
    # bundle
    #
    # prefix ex  <http://example.org/>
    #
    # prefix w3      <http://www.w3.org/>
    # prefix tr      <http://www.w3.org/TR/2011/>
    # prefix process <http://www.w3.org/2005/10/Process-20051014/tr.html#>
    # prefix email   <https://lists.w3.org/Archives/Member/w3c-archive/>
    # prefix chairs  <https://lists.w3.org/Archives/Member/chairs/>
    # prefix trans   <http://www.w3.org/2005/08/01-transitions.html#>
    # prefix rec54   <http://www.w3.org/2001/02pd/rec54#>
    #
    #
    #  entity(tr:WD-prov-dm-20111018, [ prov:type='rec54:WD' ])
    #  entity(tr:WD-prov-dm-20111215, [ prov:type='rec54:WD' ])
    #  entity(process:rec-advance,    [ prov:type='prov:Plan' ])
    #
    #
    #  entity(chairs:2011OctDec/0004, [ prov:type='trans:transreq' ])
    #  entity(email:2011Oct/0141,     [ prov:type='trans:pubreq' ])
    #  entity(email:2011Dec/0111,     [ prov:type='trans:pubreq' ])
    #
    #
    #  wasDerivedFrom(tr:WD-prov-dm-20111215, tr:WD-prov-dm-20111018)
    #
    #
    #  activity(ex:act1,-,-,[prov:type="publish"])
    #  activity(ex:act2,-,-,[prov:type="publish"])
    #
    #  wasGeneratedBy(tr:WD-prov-dm-20111018, ex:act1, -)
    #  wasGeneratedBy(tr:WD-prov-dm-20111215, ex:act2, -)
    #
    #  used(ex:act1, chairs:2011OctDec/0004, -)
    #  used(ex:act1, email:2011Oct/0141, -)
    #  used(ex:act2, email:2011Dec/0111, -)
    #
    #  agent(w3:Consortium, [ prov:type='prov:Organization' ])
    #
    #  wasAssociatedWith(ex:act1, w3:Consortium, process:rec-advance)
    #  wasAssociatedWith(ex:act2, w3:Consortium, process:rec-advance)
    #
    # endBundle
    #===========================================================================

    g = ProvDocument()
    g.add_namespace('ex', 'http://example.org/')
    g.add_namespace('w3', 'http://www.w3.org/')
    g.add_namespace('tr', 'http://www.w3.org/TR/2011/')
    g.add_namespace('process', 'http://www.w3.org/2005/10/Process-20051014/tr.html#')
    g.add_namespace('email', 'https://lists.w3.org/Archives/Member/w3c-archive/')
    g.add_namespace('chairs', 'https://lists.w3.org/Archives/Member/chairs/')
    g.add_namespace('trans', 'http://www.w3.org/2005/08/01-transitions.html#')
    g.add_namespace('rec54', 'http://www.w3.org/2001/02pd/rec54#')

    g.entity('tr:WD-prov-dm-20111018', {'prov:type': 'rec54:WD'})
    g.entity('tr:WD-prov-dm-20111215', {'prov:type': 'rec54:WD'})
    g.entity('process:rec-advance', {'prov:type': 'prov:Plan'})

    g.entity('chairs:2011OctDec/0004', {'prov:type': 'trans:transreq'})
    g.entity('email:2011Oct/0141', {'prov:type': 'trans:pubreq'})
    g.entity('email:2011Dec/0111', {'prov:type': 'trans:pubreq'})

    g.wasDerivedFrom('tr:WD-prov-dm-20111215', 'tr:WD-prov-dm-20111018')

    g.activity('ex:act1', other_attributes={'prov:type': "publish"})
    g.activity('ex:act2', other_attributes={'prov:type': "publish"})

    g.wasGeneratedBy('tr:WD-prov-dm-20111018', 'ex:act1')
    g.wasGeneratedBy('tr:WD-prov-dm-20111215', 'ex:act2')

    g.used('ex:act1', 'chairs:2011OctDec/0004')
    g.used('ex:act1', 'email:2011Oct/0141')
    g.used('ex:act2', 'email:2011Dec/0111')

    g.agent('w3:Consortium', other_attributes={'prov:type': "Organization"})

    g.wasAssociatedWith('ex:act1', 'w3:Consortium', 'process:rec-advance')
    g.wasAssociatedWith('ex:act2', 'w3:Consortium', 'process:rec-advance')

    return g


def w3c_publication_2():
    # https://github.com/lucmoreau/ProvToolbox/blob/master/asn/src/test/resources/prov/w3c-publication2.prov-asn
    #===========================================================================
    # bundle
    #
    # prefix ex <http://example.org/>
    # prefix rec <http://example.org/record>
    #
    # prefix w3 <http://www.w3.org/TR/2011/>
    # prefix hg <http://dvcs.w3.org/hg/prov/raw-file/9628aaff6e20/model/releases/WD-prov-dm-20111215/>
    #
    #
    # entity(hg:Overview.html, [ prov:type="file in hg" ])
    # entity(w3:WD-prov-dm-20111215, [ prov:type="html4" ])
    #
    #
    # activity(ex:rcp,-,-,[prov:type="copy directory"])
    #
    # wasGeneratedBy(rec:g; w3:WD-prov-dm-20111215, ex:rcp, -)
    #
    # entity(ex:req3, [ prov:type="http://www.w3.org/2005/08/01-transitions.html#pubreq" %% xsd:anyURI ])
    #
    # used(rec:u; ex:rcp,hg:Overview.html,-)
    # used(ex:rcp, ex:req3, -)
    #
    #
    # wasDerivedFrom(w3:WD-prov-dm-20111215, hg:Overview.html, ex:rcp, rec:g, rec:u)
    #
    # agent(ex:webmaster, [ prov:type='prov:Person' ])
    #
    # wasAssociatedWith(ex:rcp, ex:webmaster, -)
    #
    # endBundle
    #===========================================================================

    ex = Namespace('ex', 'http://example.org/')
    rec = Namespace('rec', 'http://example.org/record')
    w3 = Namespace('w3', 'http://www.w3.org/TR/2011/')
    hg = Namespace('hg', 'http://dvcs.w3.org/hg/prov/raw-file/9628aaff6e20/model/releases/WD-prov-dm-20111215/')

    g = ProvDocument()

    g.entity(hg['Overview.html'], {'prov:type': "file in hg"})
    g.entity(w3['WD-prov-dm-20111215'], {'prov:type': "html4"})

    g.activity(ex['rcp'], None, None, {'prov:type': "copy directory"})

    g.wasGeneratedBy('w3:WD-prov-dm-20111215', 'ex:rcp', identifier=rec['g'])

    g.entity('ex:req3', {'prov:type': Identifier("http://www.w3.org/2005/08/01-transitions.html#pubreq")})

    g.used('ex:rcp', 'hg:Overview.html', identifier='rec:u')
    g.used('ex:rcp', 'ex:req3')

    g.wasDerivedFrom('w3:WD-prov-dm-20111215', 'hg:Overview.html', 'ex:rcp', 'rec:g', 'rec:u')

    g.agent('ex:webmaster', {'prov:type': "Person"})

    g.wasAssociatedWith('ex:rcp', 'ex:webmaster')

    return g


def bundles1():
    # https://github.com/lucmoreau/ProvToolbox/blob/master/prov-n/src/test/resources/prov/bundles1.provn
    #===============================================================================
    # document
    g = ProvDocument()

    #   prefix ex  <http://example.org/example/>
    EX = Namespace("ex", "http://www.example.com/")
    g.add_namespace(EX)

    #   prefix alice  <http://example.org/alice/>
    #   prefix bob  <http://example.org/bob/>
    g.add_namespace('alice', 'http://example.org/alice/')
    g.add_namespace('bob', 'http://example.org/bob/')

    #   entity(bob:bundle1, [prov:type='prov:Bundle'])
    g.entity('bob:bundle1', {'prov:type': PROV['Bundle']})
    #   wasGeneratedBy(bob:bundle1, -, 2012-05-24T10:30:00)
    g.wasGeneratedBy('bob:bundle1', time='2012-05-24T10:30:00')
    #   agent(ex:Bob)
    g.agent('ex:Bob')
    #   wasAttributedTo(bob:bundle1, ex:Bob)
    g.wasAttributedTo('bob:bundle1', 'ex:Bob')

    #   entity(alice:bundle2, [ prov:type='prov:Bundle' ])
    g.entity('alice:bundle2', {'prov:type': PROV['Bundle']})
    #   wasGeneratedBy(alice:bundle2, -, 2012-05-25T11:15:00)
    g.wasGeneratedBy('alice:bundle2', time='2012-05-25T11:15:00')
    #   agent(ex:Alice)
    g.agent('ex:Alice')
    #   wasAttributedTo(alice:bundle2, ex:Alice)
    g.wasAttributedTo('alice:bundle2', 'ex:Alice')

    #   bundle bob:bundle1
    b1 = g.bundle('bob:bundle1')
    #     entity(ex:report1, [ prov:type="report", ex:version=1 ])
    b1.entity('ex:report1', {'prov:type': "report", 'ex:version': 1})
    #     wasGeneratedBy(ex:report1, -, 2012-05-24T10:00:01)
    b1.wasGeneratedBy('ex:report1', time='2012-05-24T10:00:01')
    #   endBundle

    #   bundle alice:bundle2
    b2 = g.bundle('alice:bundle2')
    #     entity(ex:report1)
    b2.entity('ex:report1')
    #     entity(ex:report2, [ prov:type="report", ex:version=2 ])
    b2.entity('ex:report2', {'prov:type': "report", 'ex:version': 2})
    #     wasGeneratedBy(ex:report2, -, 2012-05-25T11:00:01)
    b2.wasGeneratedBy('ex:report2', time='2012-05-25T11:00:01')
    #     wasDerivedFrom(ex:report2, ex:report1)
    b2.wasDerivedFrom('ex:report2', 'ex:report1')
    #   endBundle

    # endDocument
    return g


def bundles2():
    # https://github.com/lucmoreau/ProvToolbox/blob/master/prov-n/src/test/resources/prov/bundles2.provn
    #===========================================================================
    # document
    g = ProvDocument()

    #   prefix ex  <http://example.org/example/>
    g.add_namespace("ex", "http://www.example.com/")

    #   prefix alice  <http://example.org/alice/>
    #   prefix bob  <http://example.org/bob/>
    g.add_namespace('alice', 'http://example.org/alice/')
    g.add_namespace('bob', 'http://example.org/bob/')

    #   entity(bob:bundle4, [prov:type='prov:Bundle'])
    #   wasGeneratedBy(bob:bundle4, -, 2012-05-24T10:30:00)
    #   agent(ex:Bob)
    #   wasAttributedTo(bob:bundle4, ex:Bob)
    g.entity('bob:bundle4', {'prov:type': PROV['Bundle']})
    g.wasGeneratedBy('bob:bundle4', time='2012-05-24T10:30:00')
    g.agent('ex:Bob')
    g.wasAttributedTo('bob:bundle4', 'ex:Bob')

    #   entity(alice:bundle5, [ prov:type='prov:Bundle' ])
    #   wasGeneratedBy(alice:bundle5, -, 2012-05-25T11:15:00)
    #   agent(ex:Alice)
    #   wasAttributedTo(alice:bundle5, ex:Alice)
    g.entity('alice:bundle5', {'prov:type': PROV['Bundle']})
    g.wasGeneratedBy('alice:bundle5', time='2012-05-25T11:15:00')
    g.agent('ex:Alice')
    g.wasAttributedTo('alice:bundle5', 'ex:Alice')

    #   bundle bob:bundle4
    #     entity(ex:report1, [ prov:type="report", ex:version=1 ])
    #     wasGeneratedBy(ex:report1, -, 2012-05-24T10:00:01)
    #   endBundle
    b4 = g.bundle('bob:bundle4')
    b4.entity('ex:report1', {'prov:type': "report", 'ex:version': 1})
    b4.wasGeneratedBy('ex:report1', time='2012-05-24T10:00:01')

    #   bundle alice:bundle5
    #     entity(ex:report1bis)
    #     mentionOf(ex:report1bis, ex:report1, bob:bundle4)
    #     entity(ex:report2, [ prov:type="report", ex:version=2 ])
    #     wasGeneratedBy(ex:report2, -, 2012-05-25T11:00:01)
    #     wasDerivedFrom(ex:report2, ex:report1bis)
    #   endBundle
    b5 = g.bundle('alice:bundle5')
    b5.entity('ex:report1bis')
    b5.mentionOf('ex:report1bis', 'ex:report1', 'bob:bundle4')
    b5.entity('ex:report2', [('prov:type', "report"), ('ex:version', 2)])
    b5.wasGeneratedBy('ex:report2', time='2012-05-25T11:00:01')
    b5.wasDerivedFrom('ex:report2', 'ex:report1bis')

    # endDocument
    return g


def collections():
    g = ProvDocument()
    ex = Namespace('ex', 'http://example.org/')

    c1 = g.collection(ex['c1'])
    e1 = g.entity('ex:e1')
    g.hadMember(c1, e1)

    return g


def datatypes():
    g = ProvDocument()
    ex = Namespace('ex', 'http://example.org/')
    g.add_namespace(ex)

    attributes = {
        'ex:int': 100,
        'ex:float': 100.123456,
        'ex:long': 123456789000,
        'ex:bool': True,
        'ex:str': 'Some string',
        'ex:unicode': u'Some unicode string with accents: Huỳnh Trung Đông',
        'ex:timedate': datetime.datetime(2012, 12, 12, 14, 7, 48),
        'ex:intstr': Literal("PROV Internationalized string", PROV["InternationalizedString"], "en"),
    }
    multiline = """Line1
    Line2
Line3"""
    attributes['ex:multi-line'] = multiline
    g.entity('ex:e1', attributes)
    return g


def long_literals():
    g = ProvDocument()

    long_uri = "http://Lorem.ipsum/dolor/sit/amet/consectetur/adipiscing/elit/Quisque/vel/sollicitudin/felis/nec/" \
               "venenatis/massa/Aenean/lectus/arcu/sagittis/sit/amet/nisl/nec/varius/eleifend/sem/In/hac/habitasse/" \
               "platea/dictumst/Aliquam/eget/fermentum/enim/Curabitur/auctor/elit/non/ipsum/interdum/at/orci/aliquam/"
    ex = Namespace('ex', long_uri)
    g.add_namespace(ex)

    g.entity('ex:e1', {
        'prov:label': 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Donec pellentesque luctus nulla vel '
                      'ullamcorper. Donec sit amet ligula sit amet lorem pretium rhoncus vel vel lorem. Sed at '
                      'consequat metus, eget eleifend massa. Fusce a facilisis turpis. Lorem volutpat.'
    })

    return g

tests = [
    ('Bundle1', bundles1),
    ('Bundle2', bundles2),
    ('Primer', primer_example),
    ('W3C Publication 1', w3c_publication_1),
    ('W3C Publication 2', w3c_publication_2),
    ('collections', collections),
    ('datatypes', datatypes),
    ('Long literals', long_literals),
]



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import difflib
import glob
import inspect
import io
from lxml import etree
import os
import unittest
import warnings

from prov.identifier import Namespace, QualifiedName
from prov.constants import PROV
import prov.model as prov
from prov.tests.test_model import AllTestsBase
from prov.tests.utility import RoundTripTestCase


EX_NS = ('ex', 'http://example.com/ns/ex#')
EX_TR = ('tr', 'http://example.com/ns/tr#')

# Most general way to get the path.
DATA_PATH = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(
    inspect.currentframe()))), "xml")


def remove_empty_tags(tree):
    if tree.text is not None and tree.text.strip() == "":
        tree.text = None
    for elem in tree:
        if etree.iselement(elem):
            remove_empty_tags(elem)


def compare_xml(doc1, doc2):
    """
    Helper function to compare two XML files. It will parse both once again
    and write them in a canonical fashion.
    """
    try:
        doc1.seek(0, 0)
    except AttributeError:
        pass
    try:
        doc2.seek(0, 0)
    except AttributeError:
        pass

    obj1 = etree.parse(doc1)
    obj2 = etree.parse(doc2)

    # Remove comments from both.
    for c in obj1.getroot().xpath("//comment()"):
        p = c.getparent()
        p.remove(c)
    for c in obj2.getroot().xpath("//comment()"):
        p = c.getparent()
        p.remove(c)

    remove_empty_tags(obj1.getroot())
    remove_empty_tags(obj2.getroot())

    buf = io.BytesIO()
    obj1.write_c14n(buf)
    buf.seek(0, 0)
    str1 = buf.read().decode()
    str1 = [_i.strip() for _i in str1.splitlines() if _i.strip()]

    buf = io.BytesIO()
    obj2.write_c14n(buf)
    buf.seek(0, 0)
    str2 = buf.read().decode()
    str2 = [_i.strip() for _i in str2.splitlines() if _i.strip()]

    unified_diff = difflib.unified_diff(str1, str2)

    err_msg = "\n".join(unified_diff)
    if err_msg:
        msg = "Strings are not equal.\n"
        raise AssertionError(msg + err_msg)


class ProvXMLTestCase(unittest.TestCase):
    def test_serialization_example_6(self):
        """
        Test the serialization of example 6 which is a simple entity
        description.
        """
        document = prov.ProvDocument()
        ex_ns = document.add_namespace(*EX_NS)
        document.add_namespace(*EX_TR)

        document.entity("tr:WD-prov-dm-20111215", (
            (prov.PROV_TYPE, ex_ns["Document"]),
            ("ex:version", "2")
        ))

        with io.BytesIO() as actual:
            document.serialize(format='xml', destination=actual)
            compare_xml(os.path.join(DATA_PATH, "example_06.xml"), actual)

    def test_serialization_example_7(self):
        """
        Test the serialization of example 7 which is a basic activity.
        """
        document = prov.ProvDocument()
        document.add_namespace(*EX_NS)

        document.activity(
            "ex:a1",
            "2011-11-16T16:05:00",
            "2011-11-16T16:06:00", [
                (prov.PROV_TYPE, prov.Literal("ex:edit", prov.XSD_QNAME)),
                ("ex:host", "server.example.org")])

        with io.BytesIO() as actual:
            document.serialize(format='xml', destination=actual)
            compare_xml(os.path.join(DATA_PATH, "example_07.xml"), actual)

    def test_serialization_example_8(self):
        """
        Test the serialization of example 8 which deals with generation.
        """
        document = prov.ProvDocument()
        document.add_namespace(*EX_NS)

        e1 = document.entity("ex:e1")
        a1 = document.activity("ex:a1")

        document.wasGeneratedBy(entity=e1, activity=a1,
                                time="2001-10-26T21:32:52",
                                other_attributes={"ex:port": "p1"})

        e2 = document.entity("ex:e2")

        document.wasGeneratedBy(entity=e2, activity=a1,
                                time="2001-10-26T10:00:00",
                                other_attributes={"ex:port": "p2"})

        with io.BytesIO() as actual:
            document.serialize(format='xml', destination=actual)
            compare_xml(os.path.join(DATA_PATH, "example_08.xml"), actual)

    def test_deserialization_example_6(self):
        """
        Test the deserialization of example 6 which is a simple entity
        description.
        """
        actual_doc = prov.ProvDocument.deserialize(
            source=os.path.join(DATA_PATH, "example_06.xml"),
            format="xml")

        expected_document = prov.ProvDocument()
        ex_ns = expected_document.add_namespace(*EX_NS)
        expected_document.add_namespace(*EX_TR)

        expected_document.entity("tr:WD-prov-dm-20111215", (
            (prov.PROV_TYPE, ex_ns["Document"]),
            ("ex:version", "2")
        ))

        self.assertEqual(actual_doc, expected_document)

    def test_deserialization_example_7(self):
        """
        Test the deserialization of example 7 which is a simple activity
        description.
        """
        actual_doc = prov.ProvDocument.deserialize(
            source=os.path.join(DATA_PATH, "example_07.xml"),
            format="xml")

        expected_document = prov.ProvDocument()
        ex_ns = Namespace(*EX_NS)
        expected_document.add_namespace(ex_ns)

        expected_document.activity(
            "ex:a1",
            "2011-11-16T16:05:00",
            "2011-11-16T16:06:00", [
                (prov.PROV_TYPE, QualifiedName(ex_ns, "edit")),
                ("ex:host", "server.example.org")])

        self.assertEqual(actual_doc, expected_document)

    def test_deserialization_example_04_and_05(self):
        """
        Example 4 and 5 have a different type specification. They use an
        xsi:type as an attribute on an entity. This can be read but if
        written again it will become an XML child element. This is
        semantically identical but cannot be tested with a round trip.
        """
        # Example 4.
        xml_string = """
        <prov:document
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xmlns:xsd="http://www.w3.org/2001/XMLSchema"
            xmlns:prov="http://www.w3.org/ns/prov#"
            xmlns:ex="http://example.com/ns/ex#"
            xmlns:tr="http://example.com/ns/tr#">

          <prov:entity prov:id="tr:WD-prov-dm-20111215" xsi:type="prov:Plan">
            <prov:type xsi:type="xsd:QName">ex:Workflow</prov:type>
          </prov:entity>

        </prov:document>
        """
        with io.StringIO() as xml:
            xml.write(xml_string)
            xml.seek(0, 0)
            actual_document = prov.ProvDocument.deserialize(source=xml,
                                                            format="xml")

        expected_document = prov.ProvDocument()
        ex_ns = Namespace(*EX_NS)
        expected_document.add_namespace(ex_ns)
        expected_document.add_namespace(*EX_TR)

        # The xsi:type attribute is mapped to a proper PROV attribute.
        expected_document.entity("tr:WD-prov-dm-20111215", (
            (prov.PROV_TYPE, QualifiedName(ex_ns, "Workflow")),
            (prov.PROV_TYPE, PROV["Plan"])))

        self.assertEqual(actual_document, expected_document, "example_04")

        # Example 5.
        xml_string = """
        <prov:document
          xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
          xmlns:xsd="http://www.w3.org/2001/XMLSchema"
          xmlns:prov="http://www.w3.org/ns/prov#"
          xmlns:ex="http://example.com/ns/ex#"
          xmlns:tr="http://example.com/ns/tr#">

        <prov:entity prov:id="tr:WD-prov-dm-20111215" xsi:type="prov:Plan">
          <prov:type xsi:type="xsd:QName">ex:Workflow</prov:type>
          <prov:type xsi:type="xsd:QName">prov:Plan</prov:type> <!-- inferred -->
          <prov:type xsi:type="xsd:QName">prov:Entity</prov:type> <!-- inferred -->
        </prov:entity>

        </prov:document>
        """
        with io.StringIO() as xml:
            xml.write(xml_string)
            xml.seek(0, 0)
            actual_document = prov.ProvDocument.deserialize(source=xml,
                                                            format="xml")

        expected_document = prov.ProvDocument()
        expected_document.add_namespace(*EX_NS)
        expected_document.add_namespace(*EX_TR)

        # The xsi:type attribute is mapped to a proper PROV attribute.
        expected_document.entity("tr:WD-prov-dm-20111215", (
            (prov.PROV_TYPE, QualifiedName(ex_ns, "Workflow")),
            (prov.PROV_TYPE, PROV["Entity"]),
            (prov.PROV_TYPE, PROV["Plan"])
        ))

        self.assertEqual(actual_document, expected_document, "example_05")

    def test_other_elements(self):
        """
        PROV XML uses the <prov:other> element to enable the storage of non
        PROV information in a PROV XML document. It will be ignored by this
        library a warning will be raised informing the user.
        """
        # This is example 42 from the PROV XML documentation.
        xml_string = """
        <prov:document
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xmlns:xsd="http://www.w3.org/2001/XMLSchema"
            xmlns:prov="http://www.w3.org/ns/prov#"
            xmlns:ex="http://example.com/ns/ex#">

          <!-- prov statements go here -->

          <prov:other>
            <ex:foo>
              <ex:content>bar</ex:content>
            </ex:foo>
          </prov:other>

          <!-- more prov statements can go here -->

        </prov:document>
        """
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            with io.StringIO() as xml:
                xml.write(xml_string)
                xml.seek(0, 0)
                doc = prov.ProvDocument.deserialize(source=xml, format="xml")

        self.assertEqual(len(w), 1)
        self.assertTrue(
            "Document contains non-PROV information in <prov:other>. It will "
            "be ignored in this package." in str(w[0].message))

        # This document contains nothing else.
        self.assertEqual(len(doc._records), 0)

    def test_nested_default_namespace(self):
        """
        Tests that a default namespace that is defined in a lower level tag is
        written to a bundle.
        """
        filename = os.path.join(DATA_PATH, "nested_default_namespace.xml")
        doc = prov.ProvDocument.deserialize(source=filename, format="xml")

        ns = Namespace("", "http://example.org/0/")

        self.assertEqual(len(doc._records), 1)
        self.assertEqual(doc.get_default_namespace(), ns)
        self.assertEqual(doc._records[0].identifier.namespace, ns)
        self.assertEqual(doc._records[0].identifier.localpart, "e001")

    def test_redefining_namespaces(self):
        """
        Test the behaviour when namespaces are redefined at the element level.
        """
        filename = os.path.join(DATA_PATH,
                                "namespace_redefined_but_does_not_change.xml")
        doc = prov.ProvDocument.deserialize(source=filename, format="xml")
        # This has one record part of the original namespace.
        self.assertEqual(len(doc._records), 1)
        ns = Namespace("ex", "http://example.com/ns/ex#")
        self.assertEqual(doc._records[0].attributes[0][1].namespace, ns)

        # This also has one record but now in a different namespace.
        filename = os.path.join(DATA_PATH, "namespace_redefined.xml")
        doc = prov.ProvDocument.deserialize(source=filename, format="xml")
        new_ns = doc._records[0].attributes[0][1].namespace
        self.assertNotEqual(new_ns, ns)
        self.assertEqual(new_ns.uri, "http://example.com/ns/new_ex#")


class ProvXMLRoundTripFromFileTestCase(unittest.TestCase):
    def _perform_round_trip(self, filename, force_types=False):
        document = prov.ProvDocument.deserialize(
            source=filename, format="xml")

        with io.BytesIO() as new_xml:
            document.serialize(format='xml', destination=new_xml,
                               force_types=force_types)
            compare_xml(filename, new_xml)


# Add one test for each found file. Lazy way to do metaprogramming...
# I think parametrized tests are justified in this case as the test
# function names make it clear what is going on.
for filename in glob.iglob(os.path.join(
        DATA_PATH, "*" + os.path.extsep + "xml")):
    name = os.path.splitext(os.path.basename(filename))[0]
    test_name = "test_roundtrip_from_xml_%s" % name

    # Cannot round trip this one as the namespace in the PROV data model are
    # always defined per bundle and not per element.
    if name in ("nested_default_namespace",
                "nested_changing_default_namespace",
                "namespace_redefined_but_does_not_change",
                "namespace_redefined"):
        continue

    # Python creates closures on function calls...
    def get_fct(f):
        # Some test files have a lot of type declarations...
        if name in ["pc1"]:
            force_types = True
        else:
            force_types = False

        def fct(self):
            self._perform_round_trip(f, force_types=force_types)
        return fct

    fct = get_fct(filename)
    fct.__name__ = str(test_name)

    # Disabled round-trip XML comparisons since deserializing then serializing
    # PROV-XML does not maintain XML equivalence. (For example, prov:entity
    # elements with type prov:Plan become prov:plan elements)
    # TODO: Revisit these tests

    # setattr(ProvXMLRoundTripFromFileTestCase, test_name, fct)


class RoundTripXMLTests(RoundTripTestCase, AllTestsBase):
    FORMAT = 'xml'


if __name__ == '__main__':
    unittest.main()



"""Graphical visualisation support for prov.model.

This module produces graphical visualisation for provenanve graphs.
Requires pydot module and Graphviz.

References:

* pydot homepage: https://github.com/erocarrera/pydot
* Graphviz:       http://www.graphviz.org/
* DOT Language:   http://www.graphviz.org/doc/info/lang.html

.. moduleauthor:: Trung Dong Huynh <trungdong@donggiang.com>
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

try:
    from html import escape
except ImportError:
    from cgi import escape
from datetime import datetime
import pydot
import six

from prov.model import (
    PROV_ACTIVITY, PROV_AGENT, PROV_ALTERNATE, PROV_ASSOCIATION,
    PROV_ATTRIBUTION, PROV_BUNDLE, PROV_COMMUNICATION, PROV_DERIVATION,
    PROV_DELEGATION, PROV_ENTITY, PROV_GENERATION, PROV_INFLUENCE,
    PROV_INVALIDATION, PROV_END, PROV_MEMBERSHIP, PROV_MENTION,
    PROV_SPECIALIZATION, PROV_START, PROV_USAGE, Identifier,
    PROV_ATTRIBUTE_QNAMES, sorted_attributes, ProvException
)

__author__ = 'Trung Dong Huynh'
__email__ = 'trungdong@donggiang.com'


# Visual styles for various elements (nodes) and relations (edges)
# see http://graphviz.org/content/attrs
DOT_PROV_STYLE = {
    # Generic node
    0: {
        'shape': 'oval', 'style': 'filled',
        'fillcolor': 'lightgray', 'color': 'dimgray'
    },
    # Elements
    PROV_ENTITY: {
        'shape': 'oval', 'style': 'filled',
        'fillcolor': '#FFFC87', 'color': '#808080'
    },
    PROV_ACTIVITY: {
        'shape': 'box', 'style': 'filled',
        'fillcolor': '#9FB1FC', 'color': '#0000FF'
    },
    PROV_AGENT: {
        'shape': 'house', 'style': 'filled',
        'fillcolor': '#FED37F'
    },
    PROV_BUNDLE: {
        'shape': 'folder', 'style': 'filled',
        'fillcolor': 'aliceblue'
    },
    # Relations
    PROV_GENERATION: {
        'label': 'wasGeneratedBy', 'fontsize': '10.0',
        'color': 'darkgreen', 'fontcolor': 'darkgreen'
    },
    PROV_USAGE: {
        'label': 'used', 'fontsize': '10.0',
        'color': 'red4', 'fontcolor': 'red'
    },
    PROV_COMMUNICATION: {
        'label': 'wasInformedBy', 'fontsize': '10.0'
    },
    PROV_START: {
        'label': 'wasStartedBy', 'fontsize': '10.0'
    },
    PROV_END: {
        'label': 'wasEndedBy', 'fontsize': '10.0'
    },
    PROV_INVALIDATION: {
        'label': 'wasInvalidatedBy', 'fontsize': '10.0'
    },
    PROV_DERIVATION: {
        'label': 'wasDerivedFrom', 'fontsize': '10.0'
    },
    PROV_ATTRIBUTION: {
        'label': 'wasAttributedTo', 'fontsize': '10.0',
        'color': '#FED37F'
    },
    PROV_ASSOCIATION: {
        'label': 'wasAssociatedWith', 'fontsize': '10.0',
        'color': '#FED37F'
    },
    PROV_DELEGATION: {
        'label': 'actedOnBehalfOf', 'fontsize': '10.0',
        'color': '#FED37F'
    },
    PROV_INFLUENCE: {
        'label': 'wasInfluencedBy', 'fontsize': '10.0',
        'color': 'grey'
    },
    PROV_ALTERNATE: {
        'label': 'alternateOf', 'fontsize': '10.0'
    },
    PROV_SPECIALIZATION: {
        'label': 'specializationOf', 'fontsize': '10.0'
    },
    PROV_MENTION: {
        'label': 'mentionOf', 'fontsize': '10.0'
    },
    PROV_MEMBERSHIP: {
        'label': 'hadMember', 'fontsize': '10.0'
    },
    }

ANNOTATION_STYLE = {
    'shape': 'note', 'color': 'gray',
    'fontcolor': 'black', 'fontsize': '10'
}
ANNOTATION_LINK_STYLE = {
    'arrowhead': 'none', 'style': 'dashed',
    'color': 'gray'
}
ANNOTATION_START_ROW = '<<TABLE cellpadding=\"0\" border=\"0\">'
ANNOTATION_ROW_TEMPLATE = """    <TR>
        <TD align=\"left\" href=\"%s\">%s</TD>
        <TD align=\"left\"%s>%s</TD>
    </TR>"""
ANNOTATION_END_ROW = '    </TABLE>>'


def htlm_link_if_uri(value):
    try:
        uri = value.uri
        return '<a href="%s">%s</a>' % (uri, six.text_type(value))
    except AttributeError:
        return six.text_type(value)


def prov_to_dot(bundle, show_nary=True, use_labels=False,
                direction='BT',
                show_element_attributes=True, show_relation_attributes=True):
    """
    Convert a provenance bundle/document into a DOT graphical representation.

    :param bundle: The provenance bundle/document to be converted.
    :type bundle: :class:`ProvBundle`
    :param show_nary: shows all elements in n-ary relations.
    :type show_nary: bool
    :param use_labels: uses the prov:label property of an element as its name (instead of its identifier).
    :type use_labels: bool
    :param direction: specifies the direction of the graph. Valid values are "BT" (default), "TB", "LR", "RL".
    :param show_element_attributes: shows attributes of elements.
    :type show_element_attributes: bool
    :param show_relation_attributes: shows attributes of relations.
    :type show_relation_attributes: bool
    :returns:  :class:`pydot.Dot` -- the Dot object.
    """
    if direction not in {'BT', 'TB', 'LR', 'RL'}:
        # Invalid direction is provided
        direction = 'BT'  # reset it to the default value
    maindot = pydot.Dot(graph_type='digraph', rankdir=direction, charset='utf-8')

    node_map = {}
    count = [0, 0, 0, 0]  # counters for node ids

    def _bundle_to_dot(dot, bundle):
        def _attach_attribute_annotation(node, record):
            # Adding a node to show all attributes
            attributes = list(
                (attr_name, value) for attr_name, value in record.attributes
                if attr_name not in PROV_ATTRIBUTE_QNAMES
            )

            if not attributes:
                return  # No attribute to display

            # Sort the attributes.
            attributes = sorted_attributes(record.get_type(), attributes)

            ann_rows = [ANNOTATION_START_ROW]
            ann_rows.extend(
                ANNOTATION_ROW_TEMPLATE % (
                    attr.uri, escape(six.text_type(attr)),
                    ' href=\"%s\"' % value.uri if isinstance(value, Identifier)
                    else '',
                    escape(six.text_type(value)
                           if not isinstance(value, datetime) else
                           six.text_type(value.isoformat())))
                for attr, value in attributes
            )
            ann_rows.append(ANNOTATION_END_ROW)
            count[3] += 1
            annotations = pydot.Node(
                'ann%d' % count[3], label='\n'.join(ann_rows),
                **ANNOTATION_STYLE
            )
            dot.add_node(annotations)
            dot.add_edge(pydot.Edge(annotations, node, **ANNOTATION_LINK_STYLE))

        def _add_bundle(bundle):
            count[2] += 1
            subdot = pydot.Cluster(
                graph_name='c%d' % count[2], URL='"%s"' % bundle.identifier.uri
            )
            if use_labels:
                if bundle.label == bundle.identifier:
                    bundle_label = '"%s"' % six.text_type(bundle.label)
                else:
                    # Fancier label if both are different. The label will be
                    # the main node text, whereas the identifier will be a
                    # kind of subtitle.
                    bundle_label = ('<%s<br />'
                                    '<font color="#333333" point-size="10">'
                                    '%s</font>>')
                    bundle_label = bundle_label % (
                        six.text_type(bundle.label),
                        six.text_type(bundle.identifier)
                    )
                subdot.set_label('"%s"' % six.text_type(bundle_label))
            else:
                subdot.set_label('"%s"' % six.text_type(bundle.identifier))
            _bundle_to_dot(subdot, bundle)
            dot.add_subgraph(subdot)
            return subdot

        def _add_node(record):
            count[0] += 1
            node_id = 'n%d' % count[0]
            if use_labels:
                if record.label == record.identifier:
                    node_label = '"%s"' % six.text_type(record.label)
                else:
                    # Fancier label if both are different. The label will be
                    # the main node text, whereas the identifier will be a
                    # kind of subtitle.
                    node_label = ('<%s<br />'
                                  '<font color="#333333" point-size="10">'
                                  '%s</font>>')
                    node_label = node_label % (six.text_type(record.label),
                                               six.text_type(record.identifier))
            else:
                node_label = '"%s"' % six.text_type(record.identifier)

            uri = record.identifier.uri
            style = DOT_PROV_STYLE[record.get_type()]
            node = pydot.Node(
                node_id, label=node_label, URL='"%s"' % uri, **style
            )
            node_map[uri] = node
            dot.add_node(node)

            if show_element_attributes:
                _attach_attribute_annotation(node, rec)
            return node

        def _add_generic_node(qname):
            count[0] += 1
            node_id = 'n%d' % count[0]
            node_label = '"%s"' % six.text_type(qname)

            uri = qname.uri
            style = DOT_PROV_STYLE[0]
            node = pydot.Node(
                node_id, label=node_label, URL='"%s"' % uri, **style
            )
            node_map[uri] = node
            dot.add_node(node)
            return node

        def _get_bnode():
            count[1] += 1
            bnode_id = 'b%d' % count[1]
            bnode = pydot.Node(
                bnode_id, label='""', shape='point', color='gray'
            )
            dot.add_node(bnode)
            return bnode

        def _get_node(qname):
            if qname is None:
                return _get_bnode()
            uri = qname.uri
            if uri not in node_map:
                _add_generic_node(qname)
            return node_map[uri]

        records = bundle.get_records()
        relations = []
        for rec in records:
            if rec.is_element():
                _add_node(rec)
            else:
                # Saving the relations for later processing
                relations.append(rec)

        if not bundle.is_bundle():
            for bundle in bundle.bundles:
                _add_bundle(bundle)

        for rec in relations:
            args = rec.args
            # skipping empty records
            if not args:
                continue
            # picking element nodes
            nodes = [
                value for attr_name, value in rec.formal_attributes
                if attr_name in PROV_ATTRIBUTE_QNAMES
            ]
            other_attributes = [
                (attr_name, value) for attr_name, value in rec.attributes
                if attr_name not in PROV_ATTRIBUTE_QNAMES
            ]
            add_attribute_annotation = (
                show_relation_attributes and other_attributes
            )
            add_nary_elements = len(nodes) > 2 and show_nary
            style = DOT_PROV_STYLE[rec.get_type()]
            if len(nodes) < 2:  # too few elements for a relation?
                continue  # cannot draw this

            if add_nary_elements or add_attribute_annotation:
                # a blank node for n-ary relations or the attribute annotation
                bnode = _get_bnode()

                # the first segment
                dot.add_edge(
                    pydot.Edge(
                        _get_node(nodes[0]), bnode, arrowhead='none', **style
                    )
                )
                style = dict(style)  # copy the style
                del style['label']  # not showing label in the second segment
                # the second segment
                dot.add_edge(pydot.Edge(bnode, _get_node(nodes[1]), **style))
                if add_nary_elements:
                    style['color'] = 'gray'  # all remaining segment to be gray
                    for node in nodes[2:]:
                        if node is not None:
                            dot.add_edge(
                                pydot.Edge(bnode, _get_node(node), **style)
                            )
                if add_attribute_annotation:
                    _attach_attribute_annotation(bnode, rec)
            else:
                # show a simple binary relations with no annotation
                dot.add_edge(
                    pydot.Edge(
                        _get_node(nodes[0]), _get_node(nodes[1]), **style
                    )
                )

    try:
        unified = bundle.unified()
    except ProvException:
        # Could not unify this bundle
        # try the original document anyway
        unified = bundle

    _bundle_to_dot(maindot, unified)
    return maindot



"""
Created on Aug 13, 2015

@author: Trung Dong Huynh
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import unittest

# Skipping SVG tests if pydot is not installed
from pkgutil import find_loader
if find_loader("pydot") is not None:

    from prov.dot import prov_to_dot
    from prov.tests.test_model import AllTestsBase
    from prov.tests.utility import DocumentBaseTestCase


    class SVGDotOutputTest(DocumentBaseTestCase, AllTestsBase):
        """
        One-way output SVG with prov.dot to exercise its code
        """
        MIN_SVG_SIZE = 850

        def do_tests(self, prov_doc, msg=None):
            dot = prov_to_dot(prov_doc)
            svg_content = dot.create(format="svg")
            # Very naive check of the returned SVG content as we have no way to check the graphical content
            self.assertGreater(
                len(svg_content), self.MIN_SVG_SIZE,
                "The size of the generated SVG content should be greater than %d bytes" % self.MIN_SVG_SIZE
            )


if __name__ == '__main__':
    unittest.main()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = 'Trung Dong Huynh'
__email__ = 'trungdong@donggiang.com'
__version__ = '1.5.2'

__all__ = ["Error", "model", "read"]


class Error(Exception):
    """Base class for all errors in this package."""
    pass


def read(source, format=None):
    """
    Convenience function returning a ProvDocument instance.

    It does a lazy format detection by simply using try/except for all known
    formats. The deserializers should fail fairly early when data of the
    wrong type is passed to them thus the try/except is likely cheap. One
    could of course also do some more advanced format auto-detection but I am
    not sure that is necessary.

    The downside is that no proper error messages will be produced, use the
    format parameter to get the actual traceback.
    """
    # Lazy imports to not globber the namespace.
    from prov.model import ProvDocument

    from prov.serializers import Registry
    Registry.load_serializers()
    serializers = Registry.serializers.keys()

    if format:
        return ProvDocument.deserialize(source=source, format=format.lower())

    for format in serializers:
        try:
            return ProvDocument.deserialize(source=source, format=format)
        except:
            pass
    else:
        raise TypeError("Could not read from the source. To get a proper "
                        "error message, specify the format with the 'format' "
                        "parameter.")



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import io
import logging
import unittest

from prov.model import ProvDocument


logger = logging.getLogger(__name__)


class DocumentBaseTestCase(unittest.TestCase):
    def do_tests(self, prov_doc, msg=None):
        pass


class RoundTripTestCase(DocumentBaseTestCase):
    """A serializer test should subclass this class and set the class property
    FORMAT to the correct value (e.g. 'json', 'xml', 'rdf').
    """
    FORMAT = None  # a subclass should change this

    def do_tests(self, prov_doc, msg=None):
        self.assertRoundTripEquivalence(prov_doc, msg)

    def assertRoundTripEquivalence(self, prov_doc, msg=None):
        if self.FORMAT is None:
            # This is a dummy test, just return
            return

        with io.BytesIO() as stream:
            prov_doc.serialize(destination=stream, format=self.FORMAT, indent=4)
            stream.seek(0, 0)

            prov_doc_new = ProvDocument.deserialize(source=stream,
                                                    format=self.FORMAT)
            stream.seek(0, 0)
            # Assume UTF-8 encoding which is forced by the particular
            # PROV XML implementation and should also work for the PROV
            # JSON implementation.
            msg_extra = u"'%s' serialization content:\n%s" % (
                self.FORMAT, stream.read().decode("utf-8"))
            msg = u'\n'.join((msg, msg_extra)) if msg else msg_extra
            self.assertEqual(prov_doc, prov_doc_new, msg)



"""
Created on Jan 25, 2012

@author: Trung Dong Huynh
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import unittest
import logging
import os

from prov.model import ProvDocument, ProvBundle, ProvException, first, Literal
from prov.tests import examples
from prov.tests.attributes import TestAttributesBase
from prov.tests.qnames import TestQualifiedNamesBase
from prov.tests.statements import TestStatementsBase
from prov.tests.utility import RoundTripTestCase

logger = logging.getLogger(__name__)


EX_URI = 'http://www.example.org/'
EX2_URI = 'http://www.example2.org/'


class TestExamplesBase(object):
    """This is the base class for testing support for all the examples provided
    in prov.tests.examples.
    It is not runnable and needs to be included in a subclass of
    RoundTripTestCase.
    """
    def test_all_examples(self):
        counter = 0
        for name, graph in examples.tests:
            counter += 1
            logger.info('%d. Testing the %s example', counter, name)
            g = graph()
            self.do_tests(g)


class TestLoadingProvToolboxJSON(unittest.TestCase):
    def setUp(self):
        self.json_path = os.path.dirname(os.path.abspath(__file__)) + '/json/'
        filenames = os.listdir(self.json_path)
        self.fails = []
        for filename in filenames:
            if filename.endswith('.json'):
                with open(self.json_path + filename) as json_file:
                    try:
                        g1 = ProvDocument.deserialize(json_file)
                        json_str = g1.serialize(indent=4)
                        g2 = ProvDocument.deserialize(content=json_str)
                        self.assertEqual(
                            g1, g2,
                            'Round-trip JSON encoding/decoding failed:  %s.'
                            % filename
                        )
                    except:
                        self.fails.append(filename)

    def test_loading_all_json(self):
        # self.assertFalse(fails, 'Failed to load/round-trip %d JSON files (%s)' % (len(fails), ', '.join(fails)))

        # Code for debugging the failed tests
        for filename in self.fails:
            # Reload the failed files
            filepath = self.json_path + filename
#             os.rename(json_path + filename, json_path + filename + '-fail')
            with open(filepath) as json_file:
                logger.info("Loading %s...", filepath)
                g1 = ProvDocument.deserialize(json_file)
                json_str = g1.serialize(indent=4)
                g2 = ProvDocument.deserialize(content=json_str)
                self.assertEqual(
                    g1, g2,
                    'Round-trip JSON encoding/decoding failed:  %s.'
                    % filename
                )


class TestFlattening(unittest.TestCase):
    def test_flattening(self):
        for name, graph in examples.tests:
            logger.info('Testing flattening of the %s example', name)
            document = graph()
            flattened = document.flattened()
            flattened_records = set(flattened.get_records())
            # counting all the records:
            n_records = 0
            for record in document.get_records():
                n_records += 1
                self.assertIn(record, flattened_records)
            for bundle in document.bundles:
                for record in bundle.get_records():
                    n_records += 1
                    self.assertIn(record, flattened_records)
            self.assertEqual(n_records, len(flattened.get_records()))


class TestUnification(unittest.TestCase):
    def test_unifying(self):
        # This is a very trivial test just to exercise the unified() function
        # TODO: Create a proper unification test
        json_path = os.path.dirname(os.path.abspath(__file__)) + '/unification/'
        filenames = os.listdir(json_path)
        for filename in filenames:
            if not filename.endswith('.json'):
                continue
            filepath = json_path + filename
            with open(filepath) as json_file:
                logger.info('Testing unifying: %s', filename)
                logger.debug("Loading %s...", filepath)
                document = ProvDocument.deserialize(json_file)
                flattened = document.flattened()
                unified = flattened.unified()
                self.assertLess(
                    len(unified.get_records()),
                    len(flattened.get_records())
                )


class TestBundleUpdate(unittest.TestCase):
    def test_bundle_update_simple(self):
        doc = ProvDocument()
        doc.set_default_namespace(EX_URI)

        b1 = doc.bundle('b1')
        b1.entity('e')

        b2 = doc.bundle('b2')
        b2.entity('e')

        self.assertRaises(ProvException, lambda: b1.update(1))
        self.assertRaises(ProvException, lambda: b1.update(doc))

        b1.update(b2)
        self.assertEqual(len(b1.get_records()), 2)

    def test_document_update_simple(self):
        d1 = ProvDocument()
        d1.set_default_namespace(EX_URI)
        d1.entity('e')

        b1 = d1.bundle('b1')
        b1.entity('e')

        d2 = ProvDocument()
        d2.set_default_namespace(EX_URI)
        d2.entity('e')

        b1 = d2.bundle('b1')
        b1.entity('e')
        b2 = d2.bundle('b2')
        b2.entity('e')

        self.assertRaises(ProvException, lambda: d1.update(1))

        d1.update(d2)
        self.assertEqual(len(d1.get_records()), 2)
        self.assertEqual(len(d1.bundles), 2)


class TestAddBundle(unittest.TestCase):
    def document_1(self):
        d1 = ProvDocument()
        ns_ex = d1.add_namespace('ex', EX_URI)
        d1.entity(ns_ex['e1'])
        return d1

    def document_2(self):
        d2 = ProvDocument()
        ns_ex = d2.add_namespace('ex', EX2_URI)
        d2.activity(ns_ex['a1'])
        return d2

    def bundle_0(self):
        b = ProvBundle(namespaces={'ex': EX2_URI})
        return b

    def test_add_bundle_simple(self):
        d1 = self.document_1()
        b0 = self.bundle_0()

        def sub_test_1():
            d1.add_bundle(b0)
        self.assertRaises(ProvException, sub_test_1)
        self.assertFalse(d1.has_bundles())

        d1.add_bundle(b0, 'ex:b0')
        self.assertTrue(d1.has_bundles())
        self.assertIn(b0, d1.bundles)

        def sub_test_2():
            ex2_b0 = b0.identifier
            d1.add_bundle(ProvBundle(identifier=ex2_b0))
        self.assertRaises(ProvException, sub_test_2)

        d1.add_bundle(ProvBundle(), 'ex:b0')
        self.assertEqual(len(d1.bundles), 2)

    def test_add_bundle_document(self):
        d1 = self.document_1()
        d2 = self.document_2()

        def sub_test_1():
            d1.add_bundle(d2)
        self.assertRaises(ProvException, sub_test_1)

        ex2_b2 = d2.valid_qualified_name('ex:b2')
        d1.add_bundle(d2, 'ex:b2')
        self.assertEqual(ex2_b2, first(d1.bundles).identifier)
        self.assertNotIn(d2, d1.bundles)
        b2 = ProvBundle()
        b2.update(d2)
        self.assertIn(b2, d1.bundles)

class TestLiteralRepresentation(unittest.TestCase):
    def test_literal_provn_with_single_quotes(self):
        l = Literal('{"foo": "bar"}')
        string_rep = l.provn_representation()
        self.assertTrue('{\\"f' in string_rep)

    def test_literal_provn_with_triple_quotes(self):
        l = Literal('"""foo\\nbar"""')
        string_rep = l.provn_representation()
        self.assertTrue('\\"\\"\\"f' in string_rep)

class AllTestsBase(TestExamplesBase, TestStatementsBase,
                   TestAttributesBase, TestQualifiedNamesBase):
    """This is a test to include all available tests.
    """
    pass


class RoundTripModelTest(RoundTripTestCase, AllTestsBase):
    def assertRoundTripEquivalence(self, prov_doc, msg=None):
        """Exercises prov.model without the actual serialization and PROV-N
        generation.
        """
        provn_content = prov_doc.get_provn()
        # Checking for self-equality
        self.assertEqual(
            prov_doc, prov_doc,
            'The document is not self-equal:\n' + provn_content)


if __name__ == "__main__":
    unittest.main()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import six
from prov.identifier import Namespace

__author__ = 'Trung Dong Huynh'
__email__ = 'trungdong@donggiang.com'


XSD = Namespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
PROV = Namespace('prov', 'http://www.w3.org/ns/prov#')
XSI = Namespace('xsi', 'http://www.w3.org/2001/XMLSchema-instance')

#  C1. Entities/Activities
PROV_ENTITY = PROV['Entity']
PROV_ACTIVITY = PROV['Activity']
PROV_GENERATION = PROV['Generation']
PROV_USAGE = PROV['Usage']
PROV_COMMUNICATION = PROV['Communication']
PROV_START = PROV['Start']
PROV_END = PROV['End']
PROV_INVALIDATION = PROV['Invalidation']

#  C2. Derivations
PROV_DERIVATION = PROV['Derivation']

#  C3. Agents/Responsibility
PROV_AGENT = PROV['Agent']
PROV_ATTRIBUTION = PROV['Attribution']
PROV_ASSOCIATION = PROV['Association']
PROV_DELEGATION = PROV['Delegation']
PROV_INFLUENCE = PROV['Influence']
#  C4. Bundles
PROV_BUNDLE = PROV['Bundle']
#  C5. Alternate
PROV_ALTERNATE = PROV['Alternate']
PROV_SPECIALIZATION = PROV['Specialization']
PROV_MENTION = PROV['Mention']
#  C6. Collections
PROV_MEMBERSHIP = PROV['Membership']

PROV_N_MAP = {
    PROV_ENTITY:               u'entity',
    PROV_ACTIVITY:             u'activity',
    PROV_GENERATION:           u'wasGeneratedBy',
    PROV_USAGE:                u'used',
    PROV_COMMUNICATION:        u'wasInformedBy',
    PROV_START:                u'wasStartedBy',
    PROV_END:                  u'wasEndedBy',
    PROV_INVALIDATION:         u'wasInvalidatedBy',
    PROV_DERIVATION:           u'wasDerivedFrom',
    PROV_AGENT:                u'agent',
    PROV_ATTRIBUTION:          u'wasAttributedTo',
    PROV_ASSOCIATION:          u'wasAssociatedWith',
    PROV_DELEGATION:           u'actedOnBehalfOf',
    PROV_INFLUENCE:            u'wasInfluencedBy',
    PROV_ALTERNATE:            u'alternateOf',
    PROV_SPECIALIZATION:       u'specializationOf',
    PROV_MENTION:              u'mentionOf',
    PROV_MEMBERSHIP:           u'hadMember',
    PROV_BUNDLE:               u'bundle',
}

# Records defined as subtypes in PROV-N but top level types in for example
# PROV XML also need a mapping.
ADDITIONAL_N_MAP = {
    PROV['Revision']:          u'wasRevisionOf',
    PROV['Quotation']:         u'wasQuotedFrom',
    PROV['PrimarySource']:     u'hadPrimarySource',
    PROV['SoftwareAgent']:      u'softwareAgent',
    PROV['Person']:            u'person',
    PROV['Organization']:      u'organization',
    PROV['Plan']:              u'plan',
    PROV['Collection']:        u'collection',
    PROV['EmptyCollection']:   u'emptyCollection',
}

# Maps qualified names from the PROV namespace to their base class. If it
# has no baseclass it maps to itsself. This is needed for example for PROV
# XML (de)serializer where extended types are used a lot.
PROV_BASE_CLS = {
    PROV_ENTITY:               PROV_ENTITY,
    PROV_ACTIVITY:             PROV_ACTIVITY,
    PROV_GENERATION:           PROV_GENERATION,
    PROV_USAGE:                PROV_USAGE,
    PROV_COMMUNICATION:        PROV_COMMUNICATION,
    PROV_START:                PROV_START,
    PROV_END:                  PROV_END,
    PROV_INVALIDATION:         PROV_INVALIDATION,
    PROV_DERIVATION:           PROV_DERIVATION,
    PROV['Revision']:          PROV_DERIVATION,
    PROV['Quotation']:         PROV_DERIVATION,
    PROV['PrimarySource']:     PROV_DERIVATION,
    PROV_AGENT:                PROV_AGENT,
    PROV['SoftwareAgent']:      PROV_AGENT,
    PROV['Person']:            PROV_AGENT,
    PROV['Organization']:      PROV_AGENT,
    PROV_ATTRIBUTION:          PROV_ATTRIBUTION,
    PROV_ASSOCIATION:          PROV_ASSOCIATION,
    PROV['Plan']:              PROV_ENTITY,
    PROV_DELEGATION:           PROV_DELEGATION,
    PROV_INFLUENCE:            PROV_INFLUENCE,
    PROV_ALTERNATE:            PROV_ALTERNATE,
    PROV_SPECIALIZATION:       PROV_SPECIALIZATION,
    PROV_MENTION:              PROV_MENTION,
    PROV['Collection']:        PROV_ENTITY,
    PROV['EmptyCollection']:   PROV_ENTITY,
    PROV_MEMBERSHIP:           PROV_MEMBERSHIP,
    PROV_BUNDLE:               PROV_ENTITY
}

# Identifiers for PROV's attributes
PROV_ATTR_ENTITY = PROV['entity']
PROV_ATTR_ACTIVITY = PROV['activity']
PROV_ATTR_TRIGGER = PROV['trigger']
PROV_ATTR_INFORMED = PROV['informed']
PROV_ATTR_INFORMANT = PROV['informant']
PROV_ATTR_STARTER = PROV['starter']
PROV_ATTR_ENDER = PROV['ender']
PROV_ATTR_AGENT = PROV['agent']
PROV_ATTR_PLAN = PROV['plan']
PROV_ATTR_DELEGATE = PROV['delegate']
PROV_ATTR_RESPONSIBLE = PROV['responsible']
PROV_ATTR_GENERATED_ENTITY = PROV['generatedEntity']
PROV_ATTR_USED_ENTITY = PROV['usedEntity']
PROV_ATTR_GENERATION = PROV['generation']
PROV_ATTR_USAGE = PROV['usage']
PROV_ATTR_SPECIFIC_ENTITY = PROV['specificEntity']
PROV_ATTR_GENERAL_ENTITY = PROV['generalEntity']
PROV_ATTR_ALTERNATE1 = PROV['alternate1']
PROV_ATTR_ALTERNATE2 = PROV['alternate2']
PROV_ATTR_BUNDLE = PROV['bundle']
PROV_ATTR_INFLUENCEE = PROV['influencee']
PROV_ATTR_INFLUENCER = PROV['influencer']
PROV_ATTR_COLLECTION = PROV['collection']

#  Literal properties
PROV_ATTR_TIME = PROV['time']
PROV_ATTR_STARTTIME = PROV['startTime']
PROV_ATTR_ENDTIME = PROV['endTime']


PROV_ATTRIBUTE_QNAMES = {
    PROV_ATTR_ENTITY,
    PROV_ATTR_ACTIVITY,
    PROV_ATTR_TRIGGER,
    PROV_ATTR_INFORMED,
    PROV_ATTR_INFORMANT,
    PROV_ATTR_STARTER,
    PROV_ATTR_ENDER,
    PROV_ATTR_AGENT,
    PROV_ATTR_PLAN,
    PROV_ATTR_DELEGATE,
    PROV_ATTR_RESPONSIBLE,
    PROV_ATTR_GENERATED_ENTITY,
    PROV_ATTR_USED_ENTITY,
    PROV_ATTR_GENERATION,
    PROV_ATTR_USAGE,
    PROV_ATTR_SPECIFIC_ENTITY,
    PROV_ATTR_GENERAL_ENTITY,
    PROV_ATTR_ALTERNATE1,
    PROV_ATTR_ALTERNATE2,
    PROV_ATTR_BUNDLE,
    PROV_ATTR_INFLUENCEE,
    PROV_ATTR_INFLUENCER,
    PROV_ATTR_COLLECTION
}
PROV_ATTRIBUTE_LITERALS = {
    PROV_ATTR_TIME, PROV_ATTR_STARTTIME, PROV_ATTR_ENDTIME
}

# Set of formal attributes of PROV records
PROV_ATTRIBUTES = PROV_ATTRIBUTE_QNAMES | PROV_ATTRIBUTE_LITERALS
PROV_RECORD_ATTRIBUTES = list((attr, six.text_type(attr)) for attr in
                              PROV_ATTRIBUTES)

PROV_RECORD_IDS_MAP = dict(
    (PROV_N_MAP[rec_type_id], rec_type_id) for rec_type_id in PROV_N_MAP
)
PROV_ID_ATTRIBUTES_MAP = dict(
    (prov_id, attribute) for (prov_id, attribute) in PROV_RECORD_ATTRIBUTES
)
PROV_ATTRIBUTES_ID_MAP = dict(
    (attribute, prov_id) for (prov_id, attribute) in PROV_RECORD_ATTRIBUTES
)

# Extra definition for convenience
PROV_TYPE = PROV['type']
PROV_LABEL = PROV['label']
PROV_VALUE = PROV['value']
PROV_LOCATION = PROV['location']
PROV_ROLE = PROV['role']

PROV_QUALIFIEDNAME = PROV['QUALIFIED_NAME']

# XSD DATA TYPES
XSD_ANYURI = XSD['anyURI']
XSD_QNAME = XSD['QName']
XSD_DATETIME = XSD['dateTime']
XSD_TIME = XSD['time']
XSD_DATE = XSD['date']
XSD_STRING = XSD['string']
XSD_BOOLEAN = XSD['boolean']
# All XSD Integer types
XSD_INTEGER = XSD['integer']
XSD_LONG = XSD['long']
XSD_INT = XSD['int']
XSD_SHORT = XSD['short']
XSD_BYTE = XSD['byte']
XSD_NONNEGATIVEINTEGER = XSD['nonNegativeInteger']
XSD_UNSIGNEDLONG = XSD['unsignedLong']
XSD_UNSIGNEDINT = XSD['unsignedInt']
XSD_UNSIGNEDSHORT = XSD['unsignedShort']
XSD_UNSIGNEDBYTE = XSD['unsignedByte']
XSD_POSITIVEINTEGER = XSD['positiveInteger']
XSD_NONPOSITIVEINTEGER = XSD['nonPositiveInteger']
XSD_NEGATIVEINTEGER = XSD['negativeInteger']
# All XSD real number types
XSD_FLOAT = XSD['float']
XSD_DOUBLE = XSD['double']
XSD_DECIMAL = XSD['decimal']



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import six

__author__ = 'Trung Dong Huynh'
__email__ = 'trungdong@donggiang.com'


@six.python_2_unicode_compatible
class Identifier(object):
    """Base class for all identifiers and also represents xsd:anyURI."""
    # TODO: make Identifier an "abstract" base class and move xsd:anyURI
    # into a subclass

    def __init__(self, uri):
        """
        Constructor.

        :param uri: URI string for the long namespace identifier.
        """
        self._uri = six.text_type(uri)  # Ensure this is a unicode string

    @property
    def uri(self):
        """Identifier's URI."""
        return self._uri

    def __str__(self):
        return self._uri

    def __eq__(self, other):
        return self.uri == other.uri if isinstance(other, Identifier) else False

    def __hash__(self):
        return hash((self.uri, self.__class__))

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self._uri)

    def provn_representation(self):
        """PROV-N representation of qualified name in a string."""
        return '"%s" %%%% xsd:anyURI' % self._uri


@six.python_2_unicode_compatible
class QualifiedName(Identifier):
    """Qualified name of an identifier in a particular namespace."""

    def __init__(self, namespace, localpart):
        """
        Constructor.

        :param namespace: Namespace to use for qualified name resolution.
        :param localpart: Portion of identifier not part of the namespace prefix.
        """
        Identifier.__init__(self, u''.join([namespace.uri, localpart]))
        self._namespace = namespace
        self._localpart = localpart
        self._str = (
            ':'.join([namespace.prefix, localpart])
            if namespace.prefix else localpart
        )

    @property
    def namespace(self):
        """Namespace of qualified name."""
        return self._namespace

    @property
    def localpart(self):
        """Local part of qualified name."""
        return self._localpart

    def __str__(self):
        return self._str

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self._str)

    def __hash__(self):
        return hash(self.uri)

    def provn_representation(self):
        """PROV-N representation of qualified name in a string."""
        return "'%s'" % self._str


class Namespace(object):
    """PROV Namespace."""

    def __init__(self, prefix, uri):
        """
        Constructor.

        :param prefix: String short hand prefix for the namespace.
        :param uri: URI string for the long namespace identifier.
        """
        self._prefix = prefix
        self._uri = uri
        self._cache = dict()

    @property
    def uri(self):
        """Namespace URI."""
        return self._uri

    @property
    def prefix(self):
        """Namespace prefix."""
        return self._prefix

    def contains(self, identifier):
        """
        Indicates whether the identifier provided is contained in this namespace.

        :param identifier: Identifier to check.
        :return: bool
        """
        uri = identifier if isinstance(identifier, six.string_types) else (
            identifier.uri if isinstance(identifier, Identifier) else None
        )
        return uri.startswith(self._uri) if uri else False

    def qname(self, identifier):
        """
        Returns the qualified name of the identifier given using the namespace
        prefix.

        :param identifier: Identifier to resolve to a qualified name.
        :return: :py:class:`QualifiedName`
        """
        uri = identifier if isinstance(identifier, six.string_types) else (
            identifier.uri if isinstance(identifier, Identifier) else None
        )
        if uri and uri.startswith(self._uri):
            return QualifiedName(self, uri[len(self._uri):])
        else:
            return None

    def __eq__(self, other):
        return (
            (self._uri == other.uri and self._prefix == other.prefix)
            if isinstance(other, Namespace) else False
        )

    def __ne__(self, other):
        return (
            not isinstance(other, Namespace) or
            self._uri != other.uri or
            self._prefix != other.prefix
        )

    def __hash__(self):
        return hash((self._uri, self._prefix))

    def __repr__(self):
        return '<%s: %s {%s}>' % (
            self.__class__.__name__, self._prefix, self._uri
        )

    def __getitem__(self, localpart):
        if localpart in self._cache:
            return self._cache[localpart]
        else:
            qname = QualifiedName(self, localpart)
            self._cache[localpart] = qname
            return qname



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from prov.model import ProvDocument, Namespace


def document_with_n_bundles_having_default_namespace(n):
    prov_doc = ProvDocument()
    prov_doc.add_namespace('ex', 'http://www.example.org/')
    for i in range(n):
        x = str(i + 1)
        bundle = prov_doc.bundle('ex:bundle/' + x)
        bundle.set_default_namespace('http://www.example.org/default/' + x)
        bundle.entity('e')
    return prov_doc


class TestQualifiedNamesBase(object):
    """This is the base class for testing support for qualified names and
    namespaces. It is not runnable and needs to be included in a subclass of
    RoundTripTestCase.
    """

    def test_namespace_inheritance(self):
        prov_doc = ProvDocument()
        prov_doc.add_namespace('ex', 'http://www.example.org/')
        bundle = prov_doc.bundle('ex:bundle')
        e1 = bundle.entity('ex:e1')
        self.assertIsNotNone(e1.identifier, "e1's identifier is None!")
        self.do_tests(prov_doc)

    def test_default_namespace_inheritance(self):
        prov_doc = ProvDocument()
        prov_doc.set_default_namespace('http://www.example.org/')
        bundle = prov_doc.bundle('bundle')
        e1 = bundle.entity('e1')
        self.assertIsNotNone(e1.identifier, "e1's identifier is None!")
        self.do_tests(prov_doc)

    def test_flattening_1_bundle_with_default_namespace(self):
        prov_doc = document_with_n_bundles_having_default_namespace(1)
        flattened = prov_doc.flattened()
        self.do_tests(flattened)

    def test_flattening_2_bundles_with_default_namespace(self):
        prov_doc = document_with_n_bundles_having_default_namespace(2)
        flattened = prov_doc.flattened()
        self.do_tests(flattened)

    def test_flattening_3_bundles_with_default_namespace(self):
        prov_doc = document_with_n_bundles_having_default_namespace(3)
        flattened = prov_doc.flattened()
        self.do_tests(flattened)

    def test_flattening_1_bundle_with_default_namespaces(self):
        prov_doc = document_with_n_bundles_having_default_namespace(1)
        prov_doc.set_default_namespace('http://www.example.org/default/0')
        flattened = prov_doc.flattened()
        self.do_tests(flattened)

    def test_flattening_2_bundle_with_default_namespaces(self):
        prov_doc = document_with_n_bundles_having_default_namespace(2)
        prov_doc.set_default_namespace('http://www.example.org/default/0')
        flattened = prov_doc.flattened()
        self.do_tests(flattened)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import io
import unittest

from prov.model import *
from prov.dot import prov_to_dot
from prov.serializers import Registry
from prov.tests.examples import primer_example, primer_example_alternate


EX_NS = Namespace('ex', 'http://example.org/')
EX2_NS = Namespace('ex2', 'http://example2.org/')
EX_OTHER_NS = Namespace('other', 'http://exceptions.example.org/')


def add_label(record):
    record.add_attributes(
        [('prov:label', Literal("hello"))]
    )


def add_labels(record):
    record.add_attributes([
        ('prov:label', Literal("hello")),
        ('prov:label', Literal("bye", langtag="en")),
        ('prov:label', Literal("bonjour", langtag="fr"))
    ])


def add_types(record):
    record.add_attributes([
        ('prov:type', 'a'),
        ('prov:type', 1),
        ('prov:type', 1.0),
        ('prov:type', True),
        ('prov:type', EX_NS['abc']),
        ('prov:type', datetime.datetime.now()),
        ('prov:type', Literal('http://boiled-egg.example.com',
                              datatype=XSD_ANYURI)),
    ])


def add_locations(record):
    record.add_attributes([
        ('prov:Location', "Southampton"),
        ('prov:Location', 1),
        ('prov:Location', 1.0),
        ('prov:Location', True),
        ('prov:Location', EX_NS['london']),
        ('prov:Location', datetime.datetime.now()),
        ('prov:Location', EX_NS.uri + "london"),
        ('prov:Location', Literal(2002, datatype=XSD['gYear'])),
    ])


def add_value(record):
    record.add_attributes([
        ('prov:value', EX_NS['avalue'])
    ])


def add_further_attributes(record):
    record.add_attributes([
        (EX_NS['tag1'], "hello"),
        (EX_NS['tag2'], "bye"),
        (EX2_NS['tag3'], "hi"),
        (EX_NS['tag1'], "hello\nover\nmore\nlines"),
    ])


def add_further_attributes0(record):
    record.add_attributes([
        (EX_NS['tag1'], "hello"),
        (EX_NS['tag2'], "bye"),
        (EX_NS['tag2'], Literal("hola", langtag="es")),
        (EX2_NS['tag3'], "hi"),
        (EX_NS['tag'], 1),
        # long on python 2, int on python 3
        (EX_NS['tag'], six.integer_types[-1](1)),
        (EX_NS['tag'], Literal(1, datatype=XSD_SHORT)),
        (EX_NS['tag'], Literal(1, datatype=XSD_DOUBLE)),
        (EX_NS['tag'], 1.0),
        (EX_NS['tag'], True),
        (EX_NS['tag'], EX_NS.uri + "southampton"),
    ])

    add_further_attributes_with_qnames(record)


def add_further_attributes_with_qnames(record):
    record.add_attributes([
        (EX_NS['tag'], EX2_NS['newyork']),
        (EX_NS['tag'], EX_NS['london']),
    ])


class TestExtras(unittest.TestCase):
    def test_dot(self):
        # This is naive, since we can't programatically check the output is
        # correct
        document = ProvDocument()

        bundle1 = ProvBundle(identifier=EX_NS['bundle1'])
        bundle1.usage(
            activity=EX_NS['a1'], entity=EX_NS['e1'], identifier=EX_NS['use1']
        )
        bundle1.entity(
            identifier=EX_NS['e1'], other_attributes={PROV_ROLE: "sausage"}
        )
        bundle1.activity(identifier=EX_NS['a1'])
        document.activity(EX_NS['a2'])

        bundle2 = ProvBundle(identifier=EX_NS['bundle2'])
        bundle2.usage(
            activity=EX_NS['aa1'], entity=EX_NS['ee1'],
            identifier=EX_NS['use2']
        )
        bundle2.entity(identifier=EX_NS['ee1'])
        bundle2.activity(identifier=EX_NS['aa1'])

        document.add_bundle(bundle1)
        document.add_bundle(bundle2)
        prov_to_dot(document)

    def test_extra_attributes(self):

        document = ProvDocument()

        inf = document.influence(
            EX_NS['a2'], EX_NS['a1'], identifier=EX_NS['inf7']
        )
        add_labels(inf)
        add_types(inf)
        add_further_attributes(inf)

        self.assertEqual(
            len(inf.attributes),
            len(list(inf.formal_attributes) + inf.extra_attributes)
        )

    def test_serialize_to_path(self):
        document = ProvDocument()
        document.serialize("output.json")
        os.remove('output.json')

        document.serialize("http://netloc/outputmyprov/submit.php")

    def test_bundle_no_id(self):
        document = ProvDocument()

        def test():
            bundle = ProvBundle()
            document.add_bundle(bundle)

        self.assertRaises(ProvException, test)

    def test_use_set_time_helpers(self):
        dt = datetime.datetime.now()
        document1 = ProvDocument()
        document1.activity(EX_NS['a8'], startTime=dt, endTime=dt)

        document2 = ProvDocument()
        a = document2.activity(EX_NS['a8'])
        a.set_time(startTime=dt, endTime=dt)

        self.assertEqual(document1, document2)
        self.assertEqual(a.get_startTime(), dt)
        self.assertEqual(a.get_endTime(), dt)

    def test_bundle_add_garbage(self):
        document = ProvDocument()

        def test():
            document.add_bundle(
                document.entity(EX_NS['entity_trying_to_be_a_bundle'])
            )

        self.assertRaises(ProvException, test)

        def test():
            bundle = ProvBundle()
            document.add_bundle(bundle)

        self.assertRaises(ProvException, test)

    def test_bundle_equality_garbage(self):
        document = ProvBundle()
        self.assertNotEqual(document, 1)

    def test_bundle_is_bundle(self):
        document = ProvBundle()
        self.assertTrue(document.is_bundle())

    def test_bundle_get_record_by_id(self):
        document = ProvDocument()
        self.assertEqual(document.get_record(None), None)

        # record = document.entity(identifier=EX_NS['e1'])
        # self.assertEqual(document.get_record(EX_NS['e1']), record)
        #
        # bundle = document.bundle(EX_NS['b'])
        # self.assertEqual(bundle.get_record(EX_NS['e1']), record)

    def test_bundle_get_records(self):
        document = ProvDocument()

        document.entity(identifier=EX_NS['e1'])
        document.agent(identifier=EX_NS['e1'])
        self.assertEqual(len(list(document.get_records(ProvAgent))), 1)
        self.assertEqual(len(document.get_records()), 2)

    def test_bundle_name_clash(self):
        document = ProvDocument()

        def test():
            document.bundle(EX_NS['indistinct'])
            document.bundle(EX_NS['indistinct'])

        self.assertRaises(ProvException, test)

        document = ProvDocument()

        def test():
            document.bundle(EX_NS['indistinct'])
            bundle = ProvBundle(identifier=EX_NS['indistinct'])
            document.add_bundle(bundle)

        self.assertRaises(ProvException, test)

    def test_document_helper_methods(self):
        document = ProvDocument()
        self.assertFalse(document.is_bundle())
        self.assertFalse(document.has_bundles())
        document.bundle(EX_NS['b'])
        self.assertTrue(document.has_bundles())
        self.assertEqual(u'<ProvDocument>', str(document))

    def test_reading_and_writing_to_file_like_objects(self):
        """
        Tests reading and writing to and from file like objects.
        """
        # Create some random document.
        document = ProvDocument()
        document.entity(EX2_NS["test"])

        objects = [io.BytesIO, io.StringIO]

        Registry.load_serializers()
        formats = Registry.serializers.keys()

        for obj in objects:
            for format in formats:
                try:
                    buf = obj()
                    document.serialize(destination=buf, format=format)
                    buf.seek(0, 0)
                    new_document = ProvDocument.deserialize(source=buf,
                                                            format=format)
                    self.assertEqual(document, new_document)
                except NotImplementedError:
                    # Some serializers might not implement serialize or
                    # deserialize method
                    pass  # and this is fine in the context of this test
                finally:
                    buf.close()

    # def test_document_unification(self):
    #     # TODO: Improve testing of this...
    #     document = ProvDocument()
    #     bundle = document.bundle(identifier=EX_NS['b'])
    #     e1 = bundle.entity(EX_NS['e'])
    #     e2 = bundle.entity(EX_NS['e'])
    #     unified = document.unified()
    #
    #     self.assertEqual(len(unified._bundles[0]._records), 1)

    def test_primer_alternate(self):
        g1 = primer_example()
        g2 = primer_example_alternate()
        self.assertEqual(g1, g2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    readme = f.read()

requirements = [
    'python-dateutil>=2.2',
    'networkx>=2.0',
    'lxml>=3.3.5',
    'six>=1.9.0',
    'rdflib>=4.2.1'
]

test_requirements = [
    'pydot>=1.2.0'
]

setup(
    name='prov',
    version='1.5.2',
    description='A library for W3C Provenance Data Model supporting PROV-JSON, '
                'PROV-XML and PROV-O (RDF)',
    long_description=readme,
    author='Trung Dong Huynh',
    author_email='trungdong@donggiang.com',
    url='https://github.com/trungdong/prov',
    packages=find_packages(exclude=['dev', 'docs']),
    package_dir={
        'prov': 'prov'
    },
    scripts=['scripts/prov-convert', 'scripts/prov-compare'],
    include_package_data=True,
    install_requires=requirements,
    extras_require={
        'dot': ['pydot>=1.2.0'],
    },
    license="MIT",
    zip_safe=False,
    keywords=[
        'provenance', 'graph', 'model', 'PROV', 'PROV-DM', 'PROV-JSON', 'JSON',
        'PROV-XML', 'PROV-N', 'PROV-O', 'RDF'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Security',
        'Topic :: System :: Logging',
    ],
    test_suite='prov.tests',
    tests_require=test_requirements
)
